# # coding=utf-8
# from __future__ import unicode_literals
#
# LONG_TEXT_CONT = """
# This article's use of external links may not follow Wikipedia's policies or guidelines. Please improve this article by removing excessive and inappropriate external links. (August 2010)
# Suffix Trees by Dr. Sartaj Sahni (CISE Department Chair at University of Florida)
# Suffix Trees by Lloyd Allison
# NIST's Dictionary of Algorithms and Data Structures: Suffix Tree
# suffix_tree ANSI C implementation of a Suffix Tree
# libstree, a generic suffix tree library written in C
# Tree::Suffix, a Perl binding to libstree
# Strmat a faster generic suffix tree library written in C (uses arrays instead of linked lists)
# SuffixTree a Python binding to Strmat
# Universal Data Compression Based on the Burrows-Wheeler Transformation: Theory and Practice, application of suffix trees in the BWT
# Theory and Practice of Succinct Data Structures, C++ implementation of a compressed suffix tree]
# Practical Algorithm Template Library, a C++ library with suffix tree
#
# In computer science, a suffix tree (also called PAT tree or, in an earlier form, position tree) is a data structure that presents the suffixes of a given string in a way that allows for a particularly fast implementation of many important string operations.
# The suffix tree for a string S is a tree whose edges are labeled with strings, such that each suffix of S corresponds to exactly one path from the tree's root to a leaf. It is thus a radix tree (more specifically, a Patricia tree) for the suffixes of S.
# Constructing such a tree for the string S takes time and space linear in the length of S. Once constructed, several operations can be performed quickly, for instance locating a substring in S, locating a substring if a certain number of mistakes are allowed, locating matches for a regular expression pattern etc. Suffix trees also provided one of the first linear-time solutions for the longest common substring problem. These speedups come at a cost: storing a string's suffix tree typically requires significantly more space than storing the string itself.
# Contents [hide]
# 1 History
# 2 Definition
# 3 Generalised suffix tree
# 4 Functionality
# 5 Applications
# 6 Implementation
# 7 External construction
# 8 See also
# 9 References
# 10 External links
# [edit]History
#
# The concept was first introduced as a position tree by Weiner in 1973,[1] which Donald Knuth subsequently characterized as "Algorithm of the Year 1973". The construction was greatly simplified by McCreight in 1976 [2] , and also by Ukkonen in 1995.[3][4] Ukkonen provided the first linear-time online-construction of suffix trees, now known as Ukkonen's algorithm. These algorithms are all linear-time for constant-size alphabet, and have worst-case running time of O(nlogn) in general.
# In 1997, Farach[5] gave the first suffix tree construction algorithm that is optimal for all alphabets. In particular, this is the first linear-time algorithm for strings drawn from an alphabet of integers in a polynomial range. This latter algorithm has become the basis for new algorithms for constructing both suffix trees and suffix arrays, for example, in external memory, compressed, succinct, etc.
# [edit]Definition
#
# The suffix tree for the string S of length n is defined as a tree such that ([6] page 90):
# the paths from the root to the leaves have a one-to-one relationship with the suffixes of S,
# edges spell non-empty strings,
# and all internal nodes (except perhaps the root) have at least two children.
# Since such a tree does not exist for all strings, S is padded with a terminal symbol not seen in the string (usually denoted $). This ensures that no suffix is a prefix of another, and that there will be n leaf nodes, one for each of the n suffixes of S. Since all internal non-root nodes are branching, there can be at most n −  1 such nodes, and n + (n − 1) + 1 = 2n nodes in total (n leaves, n − 1 internal nodes, 1 root).
# Suffix links are a key feature for older linear-time construction algorithms, although most newer algorithms, which are based on Farach's algorithm, dispense with suffix links. In a complete suffix tree, all internal non-root nodes have a suffix link to another internal node. If the path from the root to a node spells the string χα, where χ is a single character and α is a string (possibly empty), it has a suffix link to the internal node representing α. See for example the suffix link from the node for ANA to the node for NA in the figure above. Suffix links are also used in some algorithms running on the tree.
#
# A suffix tree for a string S of length n can be built in Θ(n) time, if the letters come from an alphabet of integers in a polynomial range (in particular, this is true for constant-sized alphabets).[5] For larger alphabets, the running time is dominated by first sorting the letters to bring them into a range of size O(n); in general, this takes O(nlogn) time. The costs below are given under the assumption that the alphabet is constant.
# Assume that a suffix tree has been built for the string S of length n, or that a generalised suffix tree has been built for the set of strings  of total length . You can:
# Search for strings:
# Check if a string P of length m is a substring in O(m) time ([6] page 92).
# Find the first occurrence of the patterns  of total length m as substrings in O(m) time.
# Find all z occurrences of the patterns  of total length m as substrings in O(m + z) time ([6] page 123).
# Search for a regular expression P in time expected sublinear in n ([7]).
# Find for each suffix of a pattern P, the length of the longest match between a prefix of  and a substring in D in Θ(m) time ([6] page 132). This is termed the matching statistics for P.
# Find properties of the strings:
# Find the longest common substrings of the string Si and Sj in Θ(ni + nj) time ([6] page 125).
# Find all maximal pairs, maximal repeats or supermaximal repeats in Θ(n + z) time ([6] page 144).
# Find the Lempel–Ziv decomposition in Θ(n) time ([6] page 166).
# Find the longest repeated substrings in Θ(n) time.
# Find the most frequently occurring substrings of a minimum length in Θ(n) time.
# Find the shortest strings from Σ that do not occur in D, in O(n + z) time, if there are z such strings.
# Find the shortest substrings occurring only once in Θ(n) time.
# Find, for each i, the shortest substrings of Si not occurring elsewhere in D in Θ(n) time.
# The suffix tree can be prepared for constant time lowest common ancestor retrieval between nodes in Θ(n) time ([6] chapter 8). You can then also:
# Find the longest common prefix between the suffixes Si[p..ni] and Sj[q..nj] in Θ(1) ([6] page 196).
# Search for a pattern P of length m with at most k mismatches in O(kn + z) time, where z is the number of hits ([6] page 200).
# Find all z maximal palindromes in Θ(n)([6] page 198), or Θ(gn) time if gaps of length g are allowed, or Θ(kn) if k mismatches are allowed ([6] page 201).
# Find all z tandem repeats in O(nlogn + z), and k-mismatch tandem repeats in O(knlog(n / k) + z) ([6] page 204).
# Find the longest substrings common to at least k strings in D for  in Θ(n) time ([6] page 205).
# [edit]Applications
#
# Suffix trees can be used to solve a large number of string problems that occur in text-editing, free-text search, computational biology and other application areas.[8] Primary applications include:[8]
# String search, in O(m) complexity, where m is the length of the sub-string (but with initial O(n) time required to build the suffix tree for the string)
# Finding the longest repeated substring
# Finding the longest common substring
# Finding the longest palindrome in a string
# Suffix trees are often used in bioinformatics applications, searching for patterns in DNA or protein sequences (which can be viewed as long strings of characters). The ability to search efficiently with mismatches might be considered their greatest strength. Suffix trees are also used in data compression; they can be used to find repeated data, and can be used for the sorting stage of the Burrows–Wheeler transform. Variants of the LZW compression schemes use suffix trees (LZSS). A suffix tree is also used in suffix tree clustering, a data clustering algorithm used in some search engines (first introduced in [9]).
# [edit]Implementation
#
# If each node and edge can be represented in Θ(1) space, the entire tree can be represented in Θ(n) space. The total length of all the strings on all of the edges in the tree is O(n2), but each edge can be stored as the position and length of a substring of S, giving a total space usage of Θ(n) computer words. The worst-case space usage of a suffix tree is seen with a fibonacci word, giving the full 2n nodes.
# An important choice when making a suffix tree implementation is the parent-child relationships between nodes. The most common is using linked lists called sibling lists. Each node has a pointer to its first child, and to the next node in the child list it is a part of. Hash maps, sorted/unsorted arrays (with array doubling), and balanced search trees may also be used, giving different running time properties. We are interested in:
# The cost of finding the child on a given character.
# The cost of inserting a child.
# The cost of enlisting all children of a node (divided by the number of children in the table below).
# Let σ be the size of the alphabet. Then you have the following costs:
# Lookup	Insertion	Traversal
# Sibling lists / unsorted arrays	O(σ)	Θ(1)	Θ(1)
# Hash maps	Θ(1)	Θ(1)	O(σ)
# Balanced search tree	O(logσ)	O(logσ)	O(1)
# Sorted arrays	O(logσ)	O(σ)	O(1)
# Hash maps + sibling lists	O(1)	O(1)	O(1)
# Note that the insertion cost is amortised, and that the costs for hashing are given perfect hashing.
# The large amount of information in each edge and node makes the suffix tree very expensive, consuming about ten to twenty times the memory size of the source text in good implementations. The suffix array reduces this requirement to a factor of four, and researchers have continued to find smaller indexing structures.
# [edit]External construction
#
# Suffix trees quickly outgrow the main memory on standard machines for sequence collections in the order of gigabytes. As such, their construction calls for external memory approaches.
# There are theoretical results for constructing suffix trees in external memory. The algorithm by Farach et al. [10] is theoretically optimal, with an I/O complexity equal to that of sorting. However, as discussed for example in ,[11] the overall intricacy of this algorithm has prevented, so far, its practical implementation.
# On the other hand, there have been practical works for constructing disk-based suffix trees which scale to (few) GB/hours. The state of the art methods are TDD ,[12] TRELLIS [13] , DiGeST ,[14] and B2ST .[15]
# TDD and TRELLIS scale up to the entire human genome – approximately 3GB – resulting in a disk-based suffix tree of a size in the tens of gigabytes,.[12][13] However, these methods cannot handle efficiently collections of sequences exceeding 3GB.[14] DiGeST performs significantly better and is able to handle collections of sequences in the order of 6GB in about 6 hours.[14] The source code and documentation for the latter is available from [16] . All these methods can efficiently build suffix trees for the case when the tree does not fit in main memory, but the input does. The most recent method, B2ST,[15] scales to handle inputs that do not fit in main memory.
# [edit]See also
#
# Suffix array
# Generalised suffix tree
# [edit]References
#
# ^ P. Weiner (1973). "Linear pattern matching algorithm". 14th Annual IEEE Symposium on Switching and Automata Theory. pp. 1–11. doi:10.1109/SWAT.1973.13.
# ^ Edward M. McCreight (1976). "A Space-Economical Suffix Tree Construction Algorithm". Journal of the ACM 23 (2): 262–272. doi:10.1145/321941.321946.
# ^ E. Ukkonen (1995). "On-line construction of suffix trees". Algorithmica 14 (3): 249–260. doi:10.1007/BF01206331.
# ^ R. Giegerich and S. Kurtz (1997). "From Ukkonen to McCreight and Weiner: A Unifying View of Linear-Time Suffix Tree Construction". Algorithmica 19 (3): 331–353. doi:10.1007/PL00009177.
# ^ a b M. Farach (1997). "Optimal Suffix Tree Construction with Large Alphabets". FOCS: 137–143.
# ^ a b c d e f g h i j k l m n Gusfield, Dan (1999) [1997]. Algorithms on Strings, Trees and Sequences: Computer Science and Computational Biology. USA: Cambridge University Press. ISBN 0-521-58519-8.
# ^ Ricardo A. Baeza-Yates and Gaston H. Gonnet (1996). "Fast text searching for regular expressions or automaton searching on tries". Journal of the ACM (ACM Press) 43 (6): 915–936. doi:10.1145/235809.235810.
# ^ a b Allison, L.. "Suffix Trees". Retrieved 2008-10-14.
# ^ Oren Zamir and Oren Etzioni (1998). "Web document clustering: a feasibility demonstration". SIGIR '98: Proceedings of the 21st annual international ACM SIGIR conference on Research and development in information retrieval. ACM. pp. 46–54.
# ^ Martin Farach-Colton, Paolo Ferragina, S. Muthukrishnan (2000). "On the sorting-complexity of suffix tree construction.". J. Acm 47(6) 47 (6): 987–1011. doi:10.1145/355541.355547.
# ^ Smyth, William (2003). Computing Patterns in Strings. Addison-Wesley.
# ^ a b Sandeep Tata, Richard A. Hankins, and Jignesh M. Patel (2003). "Practical Suffix Tree Construction". VLDB '03: Proceedings of the 30th International Conference on Very Large Data Bases. Morgan Kaufmann. pp. 36–47.
# ^ a b Benjarath Phoophakdee and Mohammed J. Zaki (2007). "Genome-scale disk-based suffix tree indexing". SIGMOD '07: Proceedings of the ACM SIGMOD International Conference on Management of Data. ACM. pp. 833–844.
# ^ a b c Marina Barsky, Ulrike Stege, Alex Thomo, and Chris Upton (2008). "A new method for indexing genomes using on-disk suffix trees". CIKM '08: Proceedings of the 17th ACM Conference on Information and Knowledge Management. ACM. pp. 649–658.
# ^ a b Marina Barsky, Ulrike Stege, Alex Thomo, and Chris Upton (2009). "Suffix trees for very large genomic sequences". CIKM '09: Proceedings of the 18th ACM Conference on Information and Knowledge Management. ACM.
# ^ "The disk-based suffix tree for pattern search in sequenced genomes". Retrieved 2009-10-15.
# [edit]External links
#
#
# This article's use of external links may not follow Wikipedia's policies or guidelines. Please improve this article by removing excessive and inappropriate external links. (August 2010)
# Suffix Trees by Dr. Sartaj Sahni (CISE Department Chair at University of Florida)
# Suffix Trees by Lloyd Allison
# NIST's Dictionary of Algorithms and Data Structures: Suffix Tree
# suffix_tree ANSI C implementation of a Suffix Tree
# libstree, a generic suffix tree library written in C
# Tree::Suffix, a Perl binding to libstree
# Strmat a faster generic suffix tree library written in C (uses arrays instead of linked lists)
# SuffixTree a Python binding to Strmat
# Universal Data Compression Based on the Burrows-Wheeler Transformation: Theory and Practice, application of suffix trees in the BWT
# Theory and Practice of Succinct Data Structures, C++ implementation of a compressed suffix tree]
# Practical Algorithm Template Library, a C++ library with suffix treeIn computer science, a suffix tree (also called PAT tree or, in an earlier form, position tree) is a data structure that presents the suffixes of a given string in a way that allows for a particularly fast implementation of many important string operations.
# The suffix tree for a string S is a tree whose edges are labeled with strings, such that each suffix of S corresponds to exactly one path from the tree's root to a leaf. It is thus a radix tree (more specifically, a Patricia tree) for the suffixes of S.
# Constructing such a tree for the string S takes time and space linear in the length of S. Once constructed, several operations can be performed quickly, for instance locating a substring in S, locating a substring if a certain number of mistakes are allowed, locating matches for a regular expression pattern etc. Suffix trees also provided one of the first linear-time solutions for the longest common substring problem. These speedups come at a cost: storing a string's suffix tree typically requires significantly more space than storing the string itself.
# Contents [hide]
# 1 History
# 2 Definition
# 3 Generalised suffix tree
# 4 Functionality
# 5 Applications
# 6 Implementation
# 7 External construction
# 8 See also
# 9 References
# 10 External links
# [edit]History
#
# The concept was first introduced as a position tree by Weiner in 1973,[1] which Donald Knuth subsequently characterized as "Algorithm of the Year 1973". The construction was greatly simplified by McCreight in 1976 [2] , and also by Ukkonen in 1995.[3][4] Ukkonen provided the first linear-time online-construction of suffix trees, now known as Ukkonen's algorithm. These algorithms are all linear-time for constant-size alphabet, and have worst-case running time of O(nlogn) in general.
# In 1997, Farach[5] gave the first suffix tree construction algorithm that is optimal for all alphabets. In particular, this is the first linear-time algorithm for strings drawn from an alphabet of integers in a polynomial range. This latter algorithm has become the basis for new algorithms for constructing both suffix trees and suffix arrays, for example, in external memory, compressed, succinct, etc.
# [edit]Definition
#
# The suffix tree for the string S of length n is defined as a tree such that ([6] page 90):
# the paths from the root to the leaves have a one-to-one relationship with the suffixes of S,
# edges spell non-empty strings,
# and all internal nodes (except perhaps the root) have at least two children.
# Since such a tree does not exist for all strings, S is padded with a terminal symbol not seen in the string (usually denoted $). This ensures that no suffix is a prefix of another, and that there will be n leaf nodes, one for each of the n suffixes of S. Since all internal non-root nodes are branching, there can be at most n −  1 such nodes, and n + (n − 1) + 1 = 2n nodes in total (n leaves, n − 1 internal nodes, 1 root).
# Suffix links are a key feature for older linear-time construction algorithms, although most newer algorithms, which are based on Farach's algorithm, dispense with suffix links. In a complete suffix tree, all internal non-root nodes have a suffix link to another internal node. If the path from the root to a node spells the string χα, where χ is a single character and α is a string (possibly empty), it has a suffix link to the internal node representing α. See for example the suffix link from the node for ANA to the node for NA in the figure above. Suffix links are also used in some algorithms running on the tree.
#
# A suffix tree for a string S of length n can be built in Θ(n) time, if the letters come from an alphabet of integers in a polynomial range (in particular, this is true for constant-sized alphabets).[5] For larger alphabets, the running time is dominated by first sorting the letters to bring them into a range of size O(n); in general, this takes O(nlogn) time. The costs below are given under the assumption that the alphabet is constant.
# Assume that a suffix tree has been built for the string S of length n, or that a generalised suffix tree has been built for the set of strings  of total length . You can:
# Search for strings:
# Check if a string P of length m is a substring in O(m) time ([6] page 92).
# Find the first occurrence of the patterns  of total length m as substrings in O(m) time.
# Find all z occurrences of the patterns  of total length m as substrings in O(m + z) time ([6] page 123).
# Search for a regular expression P in time expected sublinear in n ([7]).
# Find for each suffix of a pattern P, the length of the longest match between a prefix of  and a substring in D in Θ(m) time ([6] page 132). This is termed the matching statistics for P.
# Find properties of the strings:
# Find the longest common substrings of the string Si and Sj in Θ(ni + nj) time ([6] page 125).
# Find all maximal pairs, maximal repeats or supermaximal repeats in Θ(n + z) time ([6] page 144).
# Find the Lempel–Ziv decomposition in Θ(n) time ([6] page 166).
# Find the longest repeated substrings in Θ(n) time.
# Find the most frequently occurring substrings of a minimum length in Θ(n) time.
# Find the shortest strings from Σ that do not occur in D, in O(n + z) time, if there are z such strings.
# Find the shortest substrings occurring only once in Θ(n) time.
# Find, for each i, the shortest substrings of Si not occurring elsewhere in D in Θ(n) time.
# The suffix tree can be prepared for constant time lowest common ancestor retrieval between nodes in Θ(n) time ([6] chapter 8). You can then also:
# Find the longest common prefix between the suffixes Si[p..ni] and Sj[q..nj] in Θ(1) ([6] page 196).
# Search for a pattern P of length m with at most k mismatches in O(kn + z) time, where z is the number of hits ([6] page 200).
# Find all z maximal palindromes in Θ(n)([6] page 198), or Θ(gn) time if gaps of length g are allowed, or Θ(kn) if k mismatches are allowed ([6] page 201).
# Find all z tandem repeats in O(nlogn + z), and k-mismatch tandem repeats in O(knlog(n / k) + z) ([6] page 204).
# Find the longest substrings common to at least k strings in D for  in Θ(n) time ([6] page 205).
# [edit]Applications
#
# Suffix trees can be used to solve a large number of string problems that occur in text-editing, free-text search, computational biology and other application areas.[8] Primary applications include:[8]
# String search, in O(m) complexity, where m is the length of the sub-string (but with initial O(n) time required to build the suffix tree for the string)
# Finding the longest repeated substring
# Finding the longest common substring
# Finding the longest palindrome in a string
# Suffix trees are often used in bioinformatics applications, searching for patterns in DNA or protein sequences (which can be viewed as long strings of characters). The ability to search efficiently with mismatches might be considered their greatest strength. Suffix trees are also used in data compression; they can be used to find repeated data, and can be used for the sorting stage of the Burrows–Wheeler transform. Variants of the LZW compression schemes use suffix trees (LZSS). A suffix tree is also used in suffix tree clustering, a data clustering algorithm used in some search engines (first introduced in [9]).
# [edit]Implementation
#
# If each node and edge can be represented in Θ(1) space, the entire tree can be represented in Θ(n) space. The total length of all the strings on all of the edges in the tree is O(n2), but each edge can be stored as the position and length of a substring of S, giving a total space usage of Θ(n) computer words. The worst-case space usage of a suffix tree is seen with a fibonacci word, giving the full 2n nodes.
# An important choice when making a suffix tree implementation is the parent-child relationships between nodes. The most common is using linked lists called sibling lists. Each node has a pointer to its first child, and to the next node in the child list it is a part of. Hash maps, sorted/unsorted arrays (with array doubling), and balanced search trees may also be used, giving different running time properties. We are interested in:
# The cost of finding the child on a given character.
# The cost of inserting a child.
# The cost of enlisting all children of a node (divided by the number of children in the table below).
# Let σ be the size of the alphabet. Then you have the following costs:
# Lookup	Insertion	Traversal
# Sibling lists / unsorted arrays	O(σ)	Θ(1)	Θ(1)
# Hash maps	Θ(1)	Θ(1)	O(σ)
# Balanced search tree	O(logσ)	O(logσ)	O(1)
# Sorted arrays	O(logσ)	O(σ)	O(1)
# Hash maps + sibling lists	O(1)	O(1)	O(1)
# Note that the insertion cost is amortised, and that the costs for hashing are given perfect hashing.
# The large amount of information in each edge and node makes the suffix tree very expensive, consuming about ten to twenty times the memory size of the source text in good implementations. The suffix array reduces this requirement to a factor of four, and researchers have continued to find smaller indexing structures.
# [edit]External construction
#
# Suffix trees quickly outgrow the main memory on standard machines for sequence collections in the order of gigabytes. As such, their construction calls for external memory approaches.
# There are theoretical results for constructing suffix trees in external memory. The algorithm by Farach et al. [10] is theoretically optimal, with an I/O complexity equal to that of sorting. However, as discussed for example in ,[11] the overall intricacy of this algorithm has prevented, so far, its practical implementation.
# On the other hand, there have been practical works for constructing disk-based suffix trees which scale to (few) GB/hours. The state of the art methods are TDD ,[12] TRELLIS [13] , DiGeST ,[14] and B2ST .[15]
# TDD and TRELLIS scale up to the entire human genome – approximately 3GB – resulting in a disk-based suffix tree of a size in the tens of gigabytes,.[12][13] However, these methods cannot handle efficiently collections of sequences exceeding 3GB.[14] DiGeST performs significantly better and is able to handle collections of sequences in the order of 6GB in about 6 hours.[14] The source code and documentation for the latter is available from [16] . All these methods can efficiently build suffix trees for the case when the tree does not fit in main memory, but the input does. The most recent method, B2ST,[15] scales to handle inputs that do not fit in main memory.
# [edit]See also
#
# Suffix array
# Generalised suffix tree
# [edit]References
#
# ^ P. Weiner (1973). "Linear pattern matching algorithm". 14th Annual IEEE Symposium on Switching and Automata Theory. pp. 1–11. doi:10.1109/SWAT.1973.13.
# ^ Edward M. McCreight (1976). "A Space-Economical Suffix Tree Construction Algorithm". Journal of the ACM 23 (2): 262–272. doi:10.1145/321941.321946.
# ^ E. Ukkonen (1995). "On-line construction of suffix trees". Algorithmica 14 (3): 249–260. doi:10.1007/BF01206331.
# ^ R. Giegerich and S. Kurtz (1997). "From Ukkonen to McCreight and Weiner: A Unifying View of Linear-Time Suffix Tree Construction". Algorithmica 19 (3): 331–353. doi:10.1007/PL00009177.
# ^ a b M. Farach (1997). "Optimal Suffix Tree Construction with Large Alphabets". FOCS: 137–143.
# ^ a b c d e f g h i j k l m n Gusfield, Dan (1999) [1997]. Algorithms on Strings, Trees and Sequences: Computer Science and Computational Biology. USA: Cambridge University Press. ISBN 0-521-58519-8.
# ^ Ricardo A. Baeza-Yates and Gaston H. Gonnet (1996). "Fast text searching for regular expressions or automaton searching on tries". Journal of the ACM (ACM Press) 43 (6): 915–936. doi:10.1145/235809.235810.
# ^ a b Allison, L.. "Suffix Trees". Retrieved 2008-10-14.
# ^ Oren Zamir and Oren Etzioni (1998). "Web document clustering: a feasibility demonstration". SIGIR '98: Proceedings of the 21st annual international ACM SIGIR conference on Research and development in information retrieval. ACM. pp. 46–54.
# ^ Martin Farach-Colton, Paolo Ferragina, S. Muthukrishnan (2000). "On the sorting-complexity of suffix tree construction.". J. Acm 47(6) 47 (6): 987–1011. doi:10.1145/355541.355547.
# ^ Smyth, William (2003). Computing Patterns in Strings. Addison-Wesley.
# ^ a b Sandeep Tata, Richard A. Hankins, and Jignesh M. Patel (2003). "Practical Suffix Tree Construction". VLDB '03: Proceedings of the 30th International Conference on Very Large Data Bases. Morgan Kaufmann. pp. 36–47.
# ^ a b Benjarath Phoophakdee and Mohammed J. Zaki (2007). "Genome-scale disk-based suffix tree indexing". SIGMOD '07: Proceedings of the ACM SIGMOD International Conference on Management of Data. ACM. pp. 833–844.
# ^ a b c Marina Barsky, Ulrike Stege, Alex Thomo, and Chris Upton (2008). "A new method for indexing genomes using on-disk suffix trees". CIKM '08: Proceedings of the 17th ACM Conference on Information and Knowledge Management. ACM. pp. 649–658.
# ^ a b Marina Barsky, Ulrike Stege, Alex Thomo, and Chris Upton (2009). "Suffix trees for very large genomic sequences". CIKM '09: Proceedings of the 18th ACM Conference on Information and Knowledge Management. ACM.
# ^ "The disk-based suffix tree for pattern search in sequenced genomes". Retrieved 2009-10-15.
# [edit]External links
# """
# LONG_TEXT = """In computer science, a suffix tree (also called PAT tree or, in an earlier form, position tree) is a data structure that presents the suffixes of a given string in a way that allows for a particularly fast implementation of many important string operations.
# The suffix tree for a string S is a tree whose edges are labeled with strings, such that each suffix of S corresponds to exactly one path from the tree's root to a leaf. It is thus a radix tree (more specifically, a Patricia tree) for the suffixes of S.
# Constructing such a tree for the string S takes time and space linear in the length of S. Once constructed, several operations can be performed quickly, for instance locating a substring in S, locating a substring if a certain number of mistakes are allowed, locating matches for a regular expression pattern etc. Suffix trees also provided one of the first linear-time solutions for the longest common substring problem. These speedups come at a cost: storing a string's suffix tree typically requires significantly more space than storing the string itself.
# Contents [hide]
# 1 History
# 2 Definition
# 3 Generalised suffix tree
# 4 Functionality
# 5 Applications
# 6 Implementation
# 7 External construction
# 8 See also
# 9 References
# 10 External links
# [edit]History
#
# The concept was first introduced as a position tree by Weiner in 1973,[1] which Donald Knuth subsequently characterized as "Algorithm of the Year 1973". The construction was greatly simplified by McCreight in 1976 [2] , and also by Ukkonen in 1995.[3][4] Ukkonen provided the first linear-time online-construction of suffix trees, now known as Ukkonen's algorithm. These algorithms are all linear-time for constant-size alphabet, and have worst-case running time of O(nlogn) in general.
# In 1997, Farach[5] gave the first suffix tree construction algorithm that is optimal for all alphabets. In particular, this is the first linear-time algorithm for strings drawn from an alphabet of integers in a polynomial range. This latter algorithm has become the basis for new algorithms for constructing both suffix trees and suffix arrays, for example, in external memory, compressed, succinct, etc.
# [edit]Definition
#
# The suffix tree for the string S of length n is defined as a tree such that ([6] page 90):
# the paths from the root to the leaves have a one-to-one relationship with the suffixes of S,
# edges spell non-empty strings,
# and all internal nodes (except perhaps the root) have at least two children.
# Since such a tree does not exist for all strings, S is padded with a terminal symbol not seen in the string (usually denoted $). This ensures that no suffix is a prefix of another, and that there will be n leaf nodes, one for each of the n suffixes of S. Since all internal non-root nodes are branching, there can be at most n −  1 such nodes, and n + (n − 1) + 1 = 2n nodes in total (n leaves, n − 1 internal nodes, 1 root).
# Suffix links are a key feature for older linear-time construction algorithms, although most newer algorithms, which are based on Farach's algorithm, dispense with suffix links. In a complete suffix tree, all internal non-root nodes have a suffix link to another internal node. If the path from the root to a node spells the string χα, where χ is a single character and α is a string (possibly empty), it has a suffix link to the internal node representing α. See for example the suffix link from the node for ANA to the node for NA in the figure above. Suffix links are also used in some algorithms running on the tree.
#
# A suffix tree for a string S of length n can be built in Θ(n) time, if the letters come from an alphabet of integers in a polynomial range (in particular, this is true for constant-sized alphabets).[5] For larger alphabets, the running time is dominated by first sorting the letters to bring them into a range of size O(n); in general, this takes O(nlogn) time. The costs below are given under the assumption that the alphabet is constant.
# Assume that a suffix tree has been built for the string S of length n, or that a generalised suffix tree has been built for the set of strings  of total length . You can:
# Search for strings:
# Check if a string P of length m is a substring in O(m) time ([6] page 92).
# Find the first occurrence of the patterns  of total length m as substrings in O(m) time.
# Find all z occurrences of the patterns  of total length m as substrings in O(m + z) time ([6] page 123).
# Search for a regular expression P in time expected sublinear in n ([7]).
# Find for each suffix of a pattern P, the length of the longest match between a prefix of  and a substring in D in Θ(m) time ([6] page 132). This is termed the matching statistics for P.
# Find properties of the strings:
# Find the longest common substrings of the string Si and Sj in Θ(ni + nj) time ([6] page 125).
# Find all maximal pairs, maximal repeats or supermaximal repeats in Θ(n + z) time ([6] page 144).
# Find the Lempel–Ziv decomposition in Θ(n) time ([6] page 166).
# Find the longest repeated substrings in Θ(n) time.
# Find the most frequently occurring substrings of a minimum length in Θ(n) time.
# Find the shortest strings from Σ that do not occur in D, in O(n + z) time, if there are z such strings.
# Find the shortest substrings occurring only once in Θ(n) time.
# Find, for each i, the shortest substrings of Si not occurring elsewhere in D in Θ(n) time.
# The suffix tree can be prepared for constant time lowest common ancestor retrieval between nodes in Θ(n) time ([6] chapter 8). You can then also:
# Find the longest common prefix between the suffixes Si[p..ni] and Sj[q..nj] in Θ(1) ([6] page 196).
# Search for a pattern P of length m with at most k mismatches in O(kn + z) time, where z is the number of hits ([6] page 200).
# Find all z maximal palindromes in Θ(n)([6] page 198), or Θ(gn) time if gaps of length g are allowed, or Θ(kn) if k mismatches are allowed ([6] page 201).
# Find all z tandem repeats in O(nlogn + z), and k-mismatch tandem repeats in O(knlog(n / k) + z) ([6] page 204).
# Find the longest substrings common to at least k strings in D for  in Θ(n) time ([6] page 205).
# [edit]Applications
#
# Suffix trees can be used to solve a large number of string problems that occur in text-editing, free-text search, computational biology and other application areas.[8] Primary applications include:[8]
# String search, in O(m) complexity, where m is the length of the sub-string (but with initial O(n) time required to build the suffix tree for the string)
# Finding the longest repeated substring
# Finding the longest common substring
# Finding the longest palindrome in a string
# Suffix trees are often used in bioinformatics applications, searching for patterns in DNA or protein sequences (which can be viewed as long strings of characters). The ability to search efficiently with mismatches might be considered their greatest strength. Suffix trees are also used in data compression; they can be used to find repeated data, and can be used for the sorting stage of the Burrows–Wheeler transform. Variants of the LZW compression schemes use suffix trees (LZSS). A suffix tree is also used in suffix tree clustering, a data clustering algorithm used in some search engines (first introduced in [9]).
# [edit]Implementation
#
# If each node and edge can be represented in Θ(1) space, the entire tree can be represented in Θ(n) space. The total length of all the strings on all of the edges in the tree is O(n2), but each edge can be stored as the position and length of a substring of S, giving a total space usage of Θ(n) computer words. The worst-case space usage of a suffix tree is seen with a fibonacci word, giving the full 2n nodes.
# An important choice when making a suffix tree implementation is the parent-child relationships between nodes. The most common is using linked lists called sibling lists. Each node has a pointer to its first child, and to the next node in the child list it is a part of. Hash maps, sorted/unsorted arrays (with array doubling), and balanced search trees may also be used, giving different running time properties. We are interested in:
# The cost of finding the child on a given character.
# The cost of inserting a child.
# The cost of enlisting all children of a node (divided by the number of children in the table below).
# Let σ be the size of the alphabet. Then you have the following costs:
# Lookup	Insertion	Traversal
# Sibling lists / unsorted arrays	O(σ)	Θ(1)	Θ(1)
# Hash maps	Θ(1)	Θ(1)	O(σ)
# Balanced search tree	O(logσ)	O(logσ)	O(1)
# Sorted arrays	O(logσ)	O(σ)	O(1)
# Hash maps + sibling lists	O(1)	O(1)	O(1)
# Note that the insertion cost is amortised, and that the costs for hashing are given perfect hashing.
# The large amount of information in each edge and node makes the suffix tree very expensive, consuming about ten to twenty times the memory size of the source text in good implementations. The suffix array reduces this requirement to a factor of four, and researchers have continued to find smaller indexing structures.
# [edit]External construction
#
# Suffix trees quickly outgrow the main memory on standard machines for sequence collections in the order of gigabytes. As such, their construction calls for external memory approaches.
# There are theoretical results for constructing suffix trees in external memory. The algorithm by Farach et al. [10] is theoretically optimal, with an I/O complexity equal to that of sorting. However, as discussed for example in ,[11] the overall intricacy of this algorithm has prevented, so far, its practical implementation.
# On the other hand, there have been practical works for constructing disk-based suffix trees which scale to (few) GB/hours. The state of the art methods are TDD ,[12] TRELLIS [13] , DiGeST ,[14] and B2ST .[15]
# TDD and TRELLIS scale up to the entire human genome – approximately 3GB – resulting in a disk-based suffix tree of a size in the tens of gigabytes,.[12][13] However, these methods cannot handle efficiently collections of sequences exceeding 3GB.[14] DiGeST performs significantly better and is able to handle collections of sequences in the order of 6GB in about 6 hours.[14] The source code and documentation for the latter is available from [16] . All these methods can efficiently build suffix trees for the case when the tree does not fit in main memory, but the input does. The most recent method, B2ST,[15] scales to handle inputs that do not fit in main memory.
# [edit]See also
#
# Suffix array
# Generalised suffix tree
# [edit]References
#
# ^ P. Weiner (1973). "Linear pattern matching algorithm". 14th Annual IEEE Symposium on Switching and Automata Theory. pp. 1–11. doi:10.1109/SWAT.1973.13.
# ^ Edward M. McCreight (1976). "A Space-Economical Suffix Tree Construction Algorithm". Journal of the ACM 23 (2): 262–272. doi:10.1145/321941.321946.
# ^ E. Ukkonen (1995). "On-line construction of suffix trees". Algorithmica 14 (3): 249–260. doi:10.1007/BF01206331.
# ^ R. Giegerich and S. Kurtz (1997). "From Ukkonen to McCreight and Weiner: A Unifying View of Linear-Time Suffix Tree Construction". Algorithmica 19 (3): 331–353. doi:10.1007/PL00009177.
# ^ a b M. Farach (1997). "Optimal Suffix Tree Construction with Large Alphabets". FOCS: 137–143.
# ^ a b c d e f g h i j k l m n Gusfield, Dan (1999) [1997]. Algorithms on Strings, Trees and Sequences: Computer Science and Computational Biology. USA: Cambridge University Press. ISBN 0-521-58519-8.
# ^ Ricardo A. Baeza-Yates and Gaston H. Gonnet (1996). "Fast text searching for regular expressions or automaton searching on tries". Journal of the ACM (ACM Press) 43 (6): 915–936. doi:10.1145/235809.235810.
# ^ a b Allison, L.. "Suffix Trees". Retrieved 2008-10-14.
# ^ Oren Zamir and Oren Etzioni (1998). "Web document clustering: a feasibility demonstration". SIGIR '98: Proceedings of the 21st annual international ACM SIGIR conference on Research and development in information retrieval. ACM. pp. 46–54.
# ^ Martin Farach-Colton, Paolo Ferragina, S. Muthukrishnan (2000). "On the sorting-complexity of suffix tree construction.". J. Acm 47(6) 47 (6): 987–1011. doi:10.1145/355541.355547.
# ^ Smyth, William (2003). Computing Patterns in Strings. Addison-Wesley.
# ^ a b Sandeep Tata, Richard A. Hankins, and Jignesh M. Patel (2003). "Practical Suffix Tree Construction". VLDB '03: Proceedings of the 30th International Conference on Very Large Data Bases. Morgan Kaufmann. pp. 36–47.
# ^ a b Benjarath Phoophakdee and Mohammed J. Zaki (2007). "Genome-scale disk-based suffix tree indexing". SIGMOD '07: Proceedings of the ACM SIGMOD International Conference on Management of Data. ACM. pp. 833–844.
# ^ a b c Marina Barsky, Ulrike Stege, Alex Thomo, and Chris Upton (2008). "A new method for indexing genomes using on-disk suffix trees". CIKM '08: Proceedings of the 17th ACM Conference on Information and Knowledge Management. ACM. pp. 649–658.
# ^ a b Marina Barsky, Ulrike Stege, Alex Thomo, and Chris Upton (2009). "Suffix trees for very large genomic sequences". CIKM '09: Proceedings of the 18th ACM Conference on Information and Knowledge Management. ACM.
# ^ "The disk-based suffix tree for pattern search in sequenced genomes". Retrieved 2009-10-15.
# [edit]External links
# """



import importlib
import itertools
import time
from abc import ABCMeta
from collections import OrderedDict
from uuid import uuid1

import six
from six import with_metaclass

from eventsourcing.exceptions import TopicResolutionError


class QualnameABCMeta(ABCMeta):
    """Supplies __qualname__ to object classes with this metaclass.
    """
    __outer_classes = {}

    if not hasattr(object, '__qualname__'):

        def __init__(cls, name, bases, dict):
            super(QualnameABCMeta, cls).__init__(name, bases, dict)
            QualnameABCMeta.__outer_classes[cls] = None
            for class_attr in dict.values():  # iterate class attributes
                for outer_class in QualnameABCMeta.__outer_classes:
                    if class_attr == outer_class:  # is the object an already registered type?
                        QualnameABCMeta.__outer_classes[outer_class] = cls
                        break

        @property
        def __qualname__(cls):
            c = QualnameABCMeta.__outer_classes[cls]
            if c is None:
                return cls.__name__
            else:
                return "%s.%s" % (c.__qualname__, cls.__name__)


def create_timesequenced_event_id():
    return uuid1().hex


class DomainEvent(with_metaclass(QualnameABCMeta)):
    """
    Base class for domain events.

    Implements methods to make instances read-only, comparable
    for equality, have recognisable representations, and hashable.
    """
    __always_encrypt__ = False

    def __init__(self, **kwargs):
        """
        Initialises event attribute values directly from constructor kwargs.
        """
        self.__dict__.update(kwargs)

    def __setattr__(self, key, value):
        """
        Inhibits event attributes from being updated by assignment.
        """
        raise AttributeError("OldDomainEvent attributes are read-only")

    def __eq__(self, rhs):
        """
        Tests for equality of type and attribute values.
        """
        if type(self) is not type(rhs):
            return NotImplemented
        return self.__dict__ == rhs.__dict__

    def __ne__(self, rhs):
        """
        Negates the equality test.
        """
        return not (self == rhs)

    def __hash__(self):
        """
        Computes a unique hash for an event, using its type and attribute values.
        """
        return hash(tuple(itertools.chain(sorted(self.__dict__.items()), [type(self)])))

    def __repr__(self):
        """
        Returns string representing the type and attribute values of the event.
        """
        return self.__class__.__qualname__ + "(" + ', '.join(
            "{0}={1!r}".format(*item) for item in sorted(self.__dict__.items())) + ')'


class EventWithEntityID(DomainEvent):
    """
    For events that have an entity ID attribute.
    """

    def __init__(self, entity_id, **kwargs):
        super(EventWithEntityID, self).__init__(**kwargs)
        self.__dict__['entity_id'] = entity_id

    @property
    def entity_id(self):
        return self.__dict__['entity_id']


class EventWithTimestamp(DomainEvent):
    """
    For events that have an timestamp attribute.
    """

    def __init__(self, timestamp=None, **kwargs):
        super(EventWithTimestamp, self).__init__(**kwargs)
        self.__dict__['timestamp'] = timestamp or time.time()

    @property
    def timestamp(self):
        return self.__dict__['timestamp']


class EventWithEntityVersion(DomainEvent):
    """
    For events that have an entity version number.
    """

    def __init__(self, entity_version, **kwargs):
        if not isinstance(entity_version, six.integer_types):
            raise TypeError("Version must be an integer: {}".format(entity_version))
        super(EventWithEntityVersion, self).__init__(**kwargs)
        self.__dict__['entity_version'] = entity_version

    @property
    def entity_version(self):
        return self.__dict__['entity_version']


class EventWithTimeuuid(DomainEvent):
    """
    For events that have an UUIDv1 event ID.
    """
    def __init__(self, event_id=None, **kwargs):
        super(EventWithTimeuuid, self).__init__(**kwargs)
        self.__dict__['event_id'] = event_id or uuid1()

    @property
    def event_id(self):
        return self.__dict__['event_id']


class TimestampedEntityEvent(EventWithTimestamp, EventWithEntityID):
    """
    For events of timestamp-based entities (e.g. a log).
    """


class VersionedEntityEvent(EventWithEntityVersion, EventWithEntityID):
    """
    For events of versioned entities.
    """


class TimeuuidedEntityEvent(EventWithTimeuuid, EventWithEntityID):
    """
    For events of entities.
    """


class TimestampedVersionedEntityEvent(EventWithTimestamp, VersionedEntityEvent):
    """
    For events of version-based entities, that are also timestamped.
    """
# class OldDomainEvent(DomainEvent):
#     """
#     Original domain event.
#
#     Due to the origins of the project, this design ended
#     up as a confusion of time-sequenced and integer-sequenced
#     events. It was used for both time-sequenced streams, such
#     as time-bucketed logs and snapshots, where the entity
#     version number appeared as the code smell known as "refused
#     bequest". It was also used for integer sequenced streams,
#     such as versioned domain entities, where the timestamp-based
#     implementation threatened inconsistency due to network issues.
#     The "solution" to the timestamp-based integer sequencing was
#     to have a second table controlling consistency of integer
#     sequenced. But working across two tables without cross-table
#     transaction support (e.g. in Cassandra) is a dead-end. Now,
#     integer sequenced domain events don't essentially require
#     to know what time they happened.
#
#     Therefore it must be much better to have two separate types
#     of domain event: time-sequenced events, and integer-sequenced
#     events. Domain entities and notification logs can have integer-
#     sequenced events, and snapshots and time-bucketed logs can
#     have time-sequenced events.
#
#     That's why this class has been deprecated :-).
#     """
#
#     def __init__(self, entity_id, entity_version, domain_event_id=None, **kwargs):
#         super(OldDomainEvent, self).__init__(**kwargs)
#         self.__dict__['entity_id'] = entity_id
#         self.__dict__['entity_version'] = entity_version
#         self.__dict__['domain_event_id'] = domain_event_id or create_timesequenced_event_id()
#
#     @property
#     def entity_id(self):
#         return self.__dict__['entity_id']
#
#     @property
#     def entity_version(self):
#         return self.__dict__['entity_version']
#
#     @property
#     def domain_event_id(self):
#         return self.__dict__['domain_event_id']
#
#     @property
#     def timestamp(self):
#         return timestamp_from_uuid(self.__dict__['domain_event_id'])


class AggregateEvent(TimestampedVersionedEntityEvent):
    """
    For events of DDD aggregates.
    """


_event_handlers = OrderedDict()


def all_events(_):
    return True


def subscribe(handler, predicate=all_events):
    if predicate not in _event_handlers:
        _event_handlers[predicate] = []
    _event_handlers[predicate].append(handler)


def unsubscribe(handler, predicate=all_events):
    if predicate in _event_handlers:
        handlers = _event_handlers[predicate]
        if handler in handlers:
            handlers.remove(handler)
            if not handlers:
                _event_handlers.pop(predicate)


def publish(event):
    matching_handlers = []
    for predicate, handlers in _event_handlers.items():
        if predicate(event):
            for handler in handlers:
                if handler not in matching_handlers:
                    matching_handlers.append(handler)
    for handler in matching_handlers:
        handler(event)


class EventHandlersNotEmptyError(Exception):
    pass


def assert_event_handlers_empty():
    if len(_event_handlers):
        msg = "Event handlers are still subscribed: %s" % _event_handlers
        raise EventHandlersNotEmptyError(msg)


def topic_from_domain_class(domain_class):
    """Returns a string describing a domain event class.

    Args:
        domain_class: A domain entity or event class.

    Returns:
        A string describing the class.
    """
    return domain_class.__module__ + '#' + getattr(domain_class, '__qualname__', domain_class.__name__)


def resolve_domain_topic(topic):
    """Return domain class described by given topic.

    Args:
        topic: A string describing a domain class.

    Returns:
        A domain class.

    Raises:
        TopicResolutionError: If there is no such domain class.
    """
    try:
        module_name, _, class_name = topic.partition('#')
        module = importlib.import_module(module_name)
    except ImportError as e:
        raise TopicResolutionError("{}: {}".format(topic, e))
    try:
        cls = resolve_attr(module, class_name)
    except AttributeError as e:
        raise TopicResolutionError("{}: {}".format(topic, e))
    return cls


def resolve_attr(obj, path):
    """A recursive version of getattr for navigating dotted paths.

    Args:
        obj: An object for which we want to retrieve a nested attribute.
        path: A dot separated string containing zero or more attribute names.

    Returns:
        The attribute referred to by obj.a1.a2.a3...

    Raises:
        AttributeError: If there is no such attribute.
    """
    if not path:
        return obj
    head, _, tail = path.partition('.')
    head_obj = getattr(obj, head)
    return resolve_attr(head_obj, tail)



import os

import cassandra.cqlengine
from cassandra import ConsistencyLevel
from cassandra.auth import PlainTextAuthProvider
from cassandra.cqlengine.management import create_keyspace_simple, drop_keyspace, sync_table

from eventsourcing.exceptions import DatasourceSettingsError
from eventsourcing.infrastructure.datastore import Datastore, DatastoreSettings

DEFAULT_HOSTS = 'localhost'
DEFAULT_PORT = 9042
DEFAULT_PROTOCOL_VERSION = 2
DEFAULT_DEFAULT_KEYSPACE = 'eventsourcing'
DEFAULT_CONSISTENCY_LEVEL = 'LOCAL_QUORUM'
DEFAULT_REPLICATION_FACTOR = 1


class CassandraSettings(DatastoreSettings):
    HOSTS = [h.strip() for h in os.getenv('CASSANDRA_HOSTS', DEFAULT_HOSTS).split(',')]
    PORT = int(os.getenv('CASSANDRA_PORT', DEFAULT_PORT))
    PROTOCOL_VERSION = int(os.getenv('CASSANDRA_PROTOCOL_VERSION', DEFAULT_PROTOCOL_VERSION))
    DEFAULT_KEYSPACE = os.getenv('CASSANDRA_KEYSPACE', DEFAULT_DEFAULT_KEYSPACE)
    CONSISTENCY_LEVEL = os.getenv('CASSANDRA_CONSISTENCY_LEVEL', DEFAULT_CONSISTENCY_LEVEL)
    REPLICATION_FACTOR = os.getenv('CASSANDRA_REPLICATION_FACTOR', DEFAULT_REPLICATION_FACTOR)

    def __init__(self, hosts=None, port=None, protocol_version=None, default_keyspace=None,
                 consistency=None, replication_factor=None, username=None, password=None):
        self.hosts = hosts or self.HOSTS
        self.port = port or self.PORT
        self.protocol_version = protocol_version or self.PROTOCOL_VERSION
        self.default_keyspace = default_keyspace or self.DEFAULT_KEYSPACE
        self.consistency = consistency or self.CONSISTENCY_LEVEL
        self.replication_factor = replication_factor or self.REPLICATION_FACTOR
        self.username = username
        self.password = password


class CassandraDatastore(Datastore):
    def setup_connection(self):
        assert isinstance(self.settings, CassandraSettings), self.settings

        # Optionally construct an "auth provider" object.
        if self.settings.username and self.settings.password:
            auth_provider = PlainTextAuthProvider(
                username=self.settings.username,
                password=self.settings.password
            )
        else:
            auth_provider = None

        # Resolve the consistency level to a driver object.
        try:
            consistency_level_name = self.settings.consistency.upper()
            consistency_level = ConsistencyLevel.name_to_value[consistency_level_name]
        except KeyError:
            msg = ("Cassandra consistency level name '{}' not found."
                   "".format(self.settings.consistency))
            raise DatasourceSettingsError(msg)

        # Use the other self.settings directly.
        cassandra.cqlengine.connection.setup(
            hosts=self.settings.hosts,
            consistency=consistency_level,
            default_keyspace=self.settings.default_keyspace,
            port=self.settings.port,
            auth_provider=auth_provider,
            protocol_version=self.settings.protocol_version,
            lazy_connect=True,
            retry_connect=True,
        )

    def drop_connection(self):
        if cassandra.cqlengine.connection.session:
            cassandra.cqlengine.connection.session.shutdown()
        if cassandra.cqlengine.connection.cluster:
            cassandra.cqlengine.connection.cluster.shutdown()

    def setup_tables(self):
        # Avoid warnings about this variable not being set.
        os.environ['CQLENG_ALLOW_SCHEMA_MANAGEMENT'] = '1'

        # Attempt to create the keyspace.
        create_keyspace_simple(
            name=self.settings.default_keyspace,
            replication_factor=self.settings.replication_factor,
        )
        for table in self.tables:
            sync_table(table)

    def drop_tables(self):
        # Avoid warnings about this variable not being set.
        os.environ['CQLENG_ALLOW_SCHEMA_MANAGEMENT'] = '1'
        drop_keyspace(name=self.settings.default_keyspace)



from eventsourcing.application.policies import CombinedPersistencePolicy
from eventsourcing.domain.model.snapshot import Snapshot
from eventsourcing.example.application import ExampleApplication
from eventsourcing.example.domainmodel import Example
from eventsourcing.example.infrastructure import ExampleRepository
from eventsourcing.infrastructure.activerecord import AbstractActiveRecordStrategy
from eventsourcing.infrastructure.eventstore import AbstractEventStore
from eventsourcing.tests.sequenced_item_tests.base import WithActiveRecordStrategies


class WithExampleApplication(WithActiveRecordStrategies):
    def construct_application(self):
        cipher = self.construct_cipher()
        app = ExampleApplication(
            integer_sequenced_active_record_strategy=self.integer_sequence_active_record_strategy,
            timestamp_sequenced_active_record_strategy=self.timestamp_sequence_active_record_strategy,
            always_encrypt=bool(cipher),
            cipher=cipher,
        )
        return app

    def construct_cipher(self):
        return None


class ExampleApplicationTestCase(WithExampleApplication):
    def test(self):
        """
        Checks the example application works in the way an example application should.
        """

        with self.construct_application() as app:
            # Check there's a stored event repo.
            self.assertIsInstance(app.integer_sequenced_active_record_strategy, AbstractActiveRecordStrategy)

            # Check there's an event store for version entity events.
            self.assertIsInstance(app.version_entity_event_store, AbstractEventStore)
            self.assertEqual(app.version_entity_event_store.active_record_strategy,
                             app.integer_sequenced_active_record_strategy)

            # Check there's an event store for timestamp entity events.
            self.assertIsInstance(app.timestamp_entity_event_store, AbstractEventStore)
            self.assertEqual(app.timestamp_entity_event_store.active_record_strategy,
                             app.timestamp_sequenced_active_record_strategy)

            # Check there's a persistence policy.
            self.assertIsInstance(app.persistence_policy, CombinedPersistencePolicy)

            # Check there's an example repository.
            self.assertIsInstance(app.example_repo, ExampleRepository)

            # Register a new example.
            example1 = app.register_new_example(a=10, b=20)
            self.assertIsInstance(example1, Example)

            # Check the example is available in the repo.
            entity1 = app.example_repo[example1.id]
            self.assertEqual(10, entity1.a)
            self.assertEqual(20, entity1.b)
            self.assertEqual(example1, entity1)

            # Change attribute values.
            entity1.a = 50

            # Check the attribute has the new value.
            self.assertEqual(50, entity1.a)

            # Check the new value is available in the repo.
            entity1 = app.example_repo[example1.id]
            self.assertEqual(50, entity1.a)

            # Take a snapshot of the entity.
            snapshot1 = app.example_repo.event_player.take_snapshot(entity1.id)

            # Take another snapshot of the entity (should be the same event).
            snapshot2 = app.example_repo.event_player.take_snapshot(entity1.id)
            self.assertEqual(snapshot1, snapshot2)

            # Check the snapshot exists.
            snapshot3 = app.example_repo.event_player.snapshot_strategy.get_snapshot(entity1.id)
            self.assertIsInstance(snapshot3, Snapshot)
            self.assertEqual(snapshot1, snapshot3)

            # Change attribute values.
            entity1.a = 100

            # Check the attribute has the new value.
            self.assertEqual(100, entity1.a)

            # Check the new value is available in the repo.
            entity1 = app.example_repo[example1.id]
            self.assertEqual(100, entity1.a)

            # Take another snapshot of the entity.
            snapshot4 = app.example_repo.event_player.take_snapshot(entity1.id)

            # Check the new snapshot is not equal to the first.
            self.assertNotEqual(snapshot1, snapshot4)



# import json
# from abc import ABCMeta, abstractmethod
#
# import requests
# import six
#
# from eventsourcing.domain.model.timebucketedlog import TimebucketedlogRepository
# from eventsourcing.domain.model.notificationlog import NotificationLog
# from eventsourcing.domain.model.sequence import SequenceRepository
# from eventsourcing.infrastructure.eventstore import AbstractEventStore
# from eventsourcing.infrastructure.notification_log import NotificationLogReader
#
#
# class ArchivedLogRepository(six.with_metaclass(ABCMeta)):
#     """
#     Provides a series of archived log documents (linked sections from the notification log).
#     """
#     @abstractmethod
#     def __getitem__(self, archived_log_id):
#         """
#         Returns archived log, for given ID.
#
#         :rtype: ArchivedLog
#
#         """
#
#
# class ArchivedLog(object):
#     def __init__(self, id, items, previous_id=None, next_id=None):
#         self.id = id
#         self.items = items
#         self.previous_id = previous_id
#         self.next_id = next_id
#
#
# class ArchivedLogRepo(ArchivedLogRepository):
#     def __init__(self, notification_log, sequence_repo, log_repo, event_store, doc_size):
#         assert isinstance(notification_log, NotificationLog), notification_log
#         assert isinstance(sequence_repo, SequenceRepository), sequence_repo
#         assert isinstance(log_repo, TimebucketedlogRepository), log_repo
#         assert isinstance(event_store, AbstractEventStore), event_store
#         assert isinstance(doc_size, six.integer_types), doc_size
#         if notification_log.sequence_size % doc_size:
#             raise ValueError("Document size {} doesn't split log sequence size {} into equal sized parts.".format(
#                 doc_size, notification_log.sequence_size
#             ))
#         self.notification_log = notification_log
#         self.sequence_repo = sequence_repo
#         self.log_repo = log_repo
#         self.event_store = event_store
#         self.doc_size = doc_size
#         self.last_last_item = None
#         self.last_slice_start = None
#         self.last_slice_stop = None
#         self.reader = NotificationLogReader(
#             notification_log=self.notification_log,
#             sequence_repo=self.sequence_repo,
#             log_repo=self.log_repo,
#             event_store=self.event_store,
#         )
#
#     def __getitem__(self, archived_log_id):
#         # Read the length of the notification log.
#         notification_log_length = len(self.reader)
#
#         # Get sequence slice start and stop indices.
#         if archived_log_id == 'current':
#             slice_start = (notification_log_length // self.doc_size) * self.doc_size
#             slice_stop = slice_start + self.doc_size
#             archived_log_id = self.format_archived_log_id(slice_start + 1, slice_stop)
#         else:
#             try:
#                 first_item_number, last_item_number = archived_log_id.split(',')
#             except ValueError as e:
#                 raise ValueError("Couldn't split archived log ID '{}': {}".format(archived_log_id, e))
#             slice_start = int(first_item_number) - 1
#             slice_stop = int(last_item_number)
#
#         if slice_start % self.doc_size:
#             raise ValueError("Document ID {} not aligned with document size {}.".format(
#                 archived_log_id, self.doc_size
#             ))
#         if slice_stop - slice_start != self.doc_size:
#             raise ValueError("Document ID {} does not match document size {}.".format(
#                 archived_log_id, self.doc_size
#             ))
#         self.last_slice_start = slice_start
#         self.last_slice_stop = slice_stop
#         items = self.reader[slice_start:slice_stop]
#
#         # Decide the IDs of previous and next archived logs.
#         if self.last_slice_start:
#             first_item_number = self.last_slice_start + 1 - self.doc_size
#             last_item_number = self.last_slice_stop - self.doc_size
#             previous_id = self.format_archived_log_id(first_item_number, last_item_number)
#         else:
#             previous_id = None
#         if self.last_slice_stop < notification_log_length:
#             first_item_number = self.last_slice_start + 1 + self.doc_size
#             last_item_number = self.last_slice_stop + self.doc_size
#             next_id = self.format_archived_log_id(first_item_number, last_item_number)
#         else:
#             next_id = None
#
#         # Return archived log object.
#         return ArchivedLog(
#             id=archived_log_id,
#             items=items,
#             previous_id=previous_id,
#             next_id=next_id,
#         )
#
#     @staticmethod
#     def format_archived_log_id(first_item_number, last_item_number):
#         return '{},{}'.format(first_item_number, last_item_number)
#
#
# class ArchivedLogReader(six.with_metaclass(ABCMeta)):
#     def __init__(self, archived_log_repo):
#         assert isinstance(archived_log_repo, ArchivedLogRepository)
#         self.archived_log_repo = archived_log_repo
#         self.archived_log_count = 0
#
#     def get_items(self, last_item_num=None):
#         self.archived_log_count = 0
#
#         # Validate the last item number.
#         if last_item_num is not None:
#             if last_item_num < 1:
#                 raise ValueError("Item number {} must be >= 1.".format(last_item_num))
#
#         # Get current doc.
#         archived_log_id = 'current'
#         archived_log = self.archived_log_repo[archived_log_id]
#
#         # Follow previous links.
#         while archived_log.previous_id:
#
#             # Break if we can go forward from here.
#             if last_item_num is not None:
#                 if int(archived_log.id.split(',')[0]) <= last_item_num + 1:
#                     break
#
#             # Get the previous document.
#             archived_log_id = archived_log.previous_id
#             archived_log = self.archived_log_repo[archived_log_id]
#
#         # Yield items in first doc, optionally after last item number.
#         items = archived_log.items
#         if last_item_num is not None:
#             doc_first_item_number = int(archived_log.id.split(',')[0])
#             from_index = last_item_num - doc_first_item_number + 1
#             items = items[from_index:]
#
#         # Yield all items in all subsequent archived logs.
#         while True:
#
#             for item in items:
#                 yield item
#             self.archived_log_count += 1
#
#             if archived_log.next_id:
#                 # Follow link to get next archived log.
#                 archived_log = self.archived_log_repo[archived_log.next_id]
#                 items = archived_log.items
#             else:
#                 break
#
#
# def deserialise_archived_log(archived_log_json):
#     try:
#         return ArchivedLog(**json.loads(archived_log_json))
#     except ValueError as e:
#         raise ValueError("Couldn't deserialise archived log: {}: {}".format(e, archived_log_json))
#
#
# def serialize_archived_log(archived_log):
#     assert isinstance(archived_log, ArchivedLog)
#     return json.dumps(archived_log.__dict__, indent=4)
#
#
# class RemoteArchivedLogRepo(ArchivedLogRepository):
#     def __init__(self, feeds_url, feed_name):
#         self.feeds_url = feeds_url
#         self.feed_name = feed_name
#
#     def __getitem__(self, archived_log_id):
#         archived_log_json = self.get_archived_log_json(archived_log_id)
#         return deserialise_archived_log(archived_log_json)
#
#     def get_archived_log_json(self, archived_log_id):
#         archived_log_url = self.make_archived_log_url(archived_log_id)
#         return self.get_resource(archived_log_url)
#
#     def get_resource(self, doc_url):
#         representation = requests.get(doc_url).content
#         if isinstance(representation, type(b'')):
#             representation = representation.decode('utf8')
#         return representation
#
#     def make_archived_log_url(self, archived_log_id):
#         return '{}/{}/{}/'.format(
#             self.feeds_url.strip('/'),
#             self.feed_name,
#             archived_log_id
#         )



from uuid import UUID

import datetime
import six

if hasattr(datetime, 'timezone'):
    utc_timezone = datetime.timezone.utc
else:
    from datetime import tzinfo, timedelta

    class UTC(tzinfo):
        def utcoffset(self, date_time):
            return timedelta(0)

        def dst(self, date_time):
            return timedelta(0)
    utc_timezone = UTC()


def timestamp_from_uuid(uuid_arg):
    """
    Return a floating point unix timestamp to 6 decimal places.

    :param uuid_arg:
    :return: Unix timestamp in seconds, with microsecond precision.
    :rtype: float
    """
    return timestamp_long_from_uuid(uuid_arg) / 1e7


def timestamp_long_from_uuid(uuid_arg):
    """
    Returns an integer value representing a unix timestamp in tenths of microseconds.

    :param uuid_arg:
    :return: Unix timestamp integer in tenths of microseconds.
    :rtype: int
    """
    return time_from_uuid(uuid_arg) - 0x01B21DD213814000


def time_from_uuid(uuid_arg):
    if isinstance(uuid_arg, six.string_types):
        uuid_arg = UUID(uuid_arg)
    assert isinstance(uuid_arg, UUID), uuid_arg
    uuid_time = uuid_arg.time
    return uuid_time



from uuid import UUID, uuid4

from eventsourcing.application.policies import PersistencePolicy
from eventsourcing.domain.model.entity import TimeuuidedVersionedEntity
from eventsourcing.domain.model.events import EventWithTimeuuid, publish
from eventsourcing.infrastructure.cassandra.activerecords import CassandraActiveRecordStrategy, \
    CqlTimeuuidSequencedItem
from eventsourcing.infrastructure.cassandra.datastore import CassandraDatastore, CassandraSettings
from eventsourcing.infrastructure.eventsourcedrepository import EventSourcedRepository
from eventsourcing.infrastructure.eventstore import EventStore
from eventsourcing.infrastructure.sequenceditem import SequencedItem
from eventsourcing.infrastructure.sequenceditemmapper import SequencedItemMapper
from eventsourcing.tests.datastore_tests.base import AbstractDatastoreTestCase
from eventsourcing.tests.datastore_tests.test_cassandra import DEFAULT_KEYSPACE_FOR_TESTING
from eventsourcing.utils.time import timestamp_from_uuid

# This test has events with TimeUUID value as the 'event ID'. How easy is it to customize
# the infrastructure to support that? We just need to make a model that uses these events,
# define a suitable database table, and configure the other components. It's easy.

# Firstly, define and entity that uses events with TimeUUIDs.
class ExampleEntity(TimeuuidedVersionedEntity):

    def __init__(self, **kwargs):
        super(ExampleEntity, self).__init__(**kwargs)
        self._is_finished = False

    class Started(EventWithTimeuuid):
        pass

    class Finished(EventWithTimeuuid):
        pass

    def finish(self):
        event = ExampleEntity.Finished(
            entity_id=self.id,
            entity_version=self.version,
        )
        self._apply(event)
        publish(event)

    @staticmethod
    def _mutator(event, initial):
        if isinstance(event, ExampleEntity.Started):
            return ExampleEntity(**event.__dict__)
        elif isinstance(event, ExampleEntity.Finished):
            initial._is_finished = True
            return initial

    @classmethod
    def start(cls):
        event = ExampleEntity.Started(entity_id=uuid4())
        entity = ExampleEntity.mutate(None, event)
        publish(event)
        return entity


# Define a suitable active record class.
#  - class CqlTimeuuidSequencedItem has been moved into the library

# Define an application that uses the entity class and the table
class ExampleApplicationWithTimeuuidSequencedItems(object):
    def __init__(self):
        self.event_store = EventStore(
            active_record_strategy=CassandraActiveRecordStrategy(
                active_record_class=CqlTimeuuidSequencedItem,
                sequenced_item_class=SequencedItem,
            ),
            sequenced_item_mapper=SequencedItemMapper(
                sequenced_item_class=SequencedItem,
                event_sequence_id_attr='entity_id',
                event_position_attr='event_id',
            )
        )
        self.repository = EventSourcedRepository(
            mutator=ExampleEntity.mutate,
            event_store=self.event_store,
        )
        self.persistence_policy = PersistencePolicy(self.event_store)

    def start_entity(self):
        return ExampleEntity.start()

    def close(self):
        self.persistence_policy.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class TestDomainEventsWithTimeUUIDs(AbstractDatastoreTestCase):
    def setUp(self):
        super(TestDomainEventsWithTimeUUIDs, self).setUp()
        self.datastore.setup_connection()
        self.datastore.setup_tables()

    def tearDown(self):
        self.datastore.drop_tables()
        self.datastore.drop_connection()
        super(TestDomainEventsWithTimeUUIDs, self).setUp()

    def construct_datastore(self):
        return CassandraDatastore(
            settings=CassandraSettings(default_keyspace=DEFAULT_KEYSPACE_FOR_TESTING),
            tables=(CqlTimeuuidSequencedItem,),
        )

    def test(self):
        with ExampleApplicationWithTimeuuidSequencedItems() as app:

            # Create entity.
            entity1 = app.start_entity()
            self.assertIsInstance(entity1._initial_event_id, UUID)
            self.assertEqual(entity1.created_on, timestamp_from_uuid(entity1._initial_event_id))

            # Read entity from repo.
            retrieved_obj = app.repository[entity1.id]
            self.assertEqual(retrieved_obj.id, entity1.id)



import six
from cassandra.cqlengine.functions import Token
from cassandra.cqlengine.models import Model, columns
from cassandra.cqlengine.query import LWTException, BatchQuery

from eventsourcing.infrastructure.activerecord import AbstractActiveRecordStrategy


class CassandraActiveRecordStrategy(AbstractActiveRecordStrategy):

    def append_item(self, sequenced_item):
        if isinstance(sequenced_item, list):
            with BatchQuery() as b:
                for i in sequenced_item:
                    assert isinstance(i, self.sequenced_item_class), (type(i), self.sequenced_item_class)
                    self.active_record_class.batch(b).create(
                        s=i.sequence_id,
                        p=i.position,
                        t=i.topic,
                        d=i.data,
                    )
        else:
            active_record = self.to_active_record(sequenced_item)
            try:
                active_record.save()
            except LWTException as e:
                self.raise_sequenced_item_error(sequenced_item, e)

    def get_item(self, sequence_id, eq):
        query = self.filter(s=sequence_id, p__eq=eq)
        items = six.moves.map(self.from_active_record, query)
        items = list(items)
        try:
            return items[0]
        except IndexError:
            self.raise_index_error(eq)

    def get_items(self, sequence_id, gt=None, gte=None, lt=None, lte=None, limit=None,
                  query_ascending=True, results_ascending=True):

        assert limit is None or limit >= 1, limit
        assert not (gte and gt)
        assert not (lte and lt)

        query = self.filter(s=sequence_id)

        if query_ascending:
            query = query.order_by('p')

        if gt is not None:
            query = query.filter(p__gt=gt)
        if gte is not None:
            query = query.filter(p__gte=gte)
        if lt is not None:
            query = query.filter(p__lt=lt)
        if lte is not None:
            query = query.filter(p__lte=lte)

        if limit is not None:
            query = query.limit(limit)

        items = six.moves.map(self.from_active_record, query)

        items = list(items)

        if results_ascending != query_ascending:
            items.reverse()

        return items

    def all_items(self):
        query = self.active_record_class.objects.all().limit(10)
        page = list(query)
        while page:
            for record in page:
                yield self.from_active_record(record)
            last = page[-1]
            page = list(query.filter(pk__token__gt=Token(last.pk)))

    def to_active_record(self, sequenced_item):
        """
        Returns an active record instance, from given sequenced item.
        """
        if isinstance(sequenced_item, list):
            return [self.to_active_record(i) for i in sequenced_item]
        assert isinstance(sequenced_item, self.sequenced_item_class), (type(sequenced_item), self.sequenced_item_class)
        return self.active_record_class(
            s=sequenced_item.sequence_id,
            p=sequenced_item.position,
            t=sequenced_item.topic,
            d=sequenced_item.data
        )

    def from_active_record(self, active_record):
        """
        Returns a sequenced item instance, from given active record.
        """
        return self.sequenced_item_class(
            sequence_id=active_record.s,
            position=active_record.p,
            topic=active_record.t,
            data=active_record.d,
        )

    def filter(self, *args, **kwargs):
        return self.active_record_class.objects.filter(*args, **kwargs)


class CqlIntegerSequencedItem(Model):
    """Stores integer-sequenced items in Cassandra."""

    _if_not_exists = True

    __table_name__ = 'integer_sequenced_items'

    # Sequence ID (e.g. an entity or aggregate ID).
    s = columns.UUID(partition_key=True)

    # Position (index) of item in sequence.
    p = columns.BigInt(clustering_order='DESC', primary_key=True)

    # Topic of the item (e.g. path to domain event class).
    t = columns.Text(required=True)

    # State of the item (serialized dict, possibly encrypted).
    d = columns.Text(required=True)


class CqlTimestampSequencedItem(Model):
    """Stores timestamp-sequenced items in Cassandra."""

    _if_not_exists = True

    __table_name__ = 'timestamp_sequenced_items'

    # Sequence ID (e.g. an entity or aggregate ID).
    s = columns.UUID(partition_key=True)

    # Position (in time) of item in sequence.
    p = columns.Double(clustering_order='DESC', primary_key=True)

    # Topic of the item (e.g. path to domain event class).
    t = columns.Text(required=True)

    # State of the item (serialized dict, possibly encrypted).
    d = columns.Text(required=True)


class CqlTimeuuidSequencedItem(Model):
    """Stores timeuuid-sequenced items in Cassandra."""

    _if_not_exists = True

    __table_name__ = 'timeuuid_sequenced_items'

    # Sequence UUID (e.g. an entity or aggregate ID).
    s = columns.UUID(partition_key=True)

    # Position (in time) of item in sequence.
    p = columns.TimeUUID(clustering_order='DESC', primary_key=True)

    # Topic of the item (e.g. path to domain event class).
    t = columns.Text(required=True)

    # State of the item (serialized dict, possibly encrypted).
    d = columns.Text(required=True)



from abc import abstractmethod

from eventsourcing.infrastructure.datastore import DatastoreConnectionError, DatastoreTableError
from eventsourcing.tests.base import AbstractTestCase


class AbstractDatastoreTestCase(AbstractTestCase):

    def __init__(self, *args, **kwargs):
        super(AbstractDatastoreTestCase, self).__init__(*args, **kwargs)
        self._datastore = None

    def tearDown(self):
        self._datastore = None
        super(AbstractDatastoreTestCase, self).tearDown()

    @property
    def datastore(self):
        """
        :rtype: eventsourcing.infrastructure.datastore.datastore.Datastore
        """
        if self._datastore is None:
            self._datastore = self.construct_datastore()
        return self._datastore

    @abstractmethod
    def construct_datastore(self):
        """
        :rtype: eventsourcing.infrastructure.datastore.datastore.Datastore
        """


class DatastoreTestCase(AbstractDatastoreTestCase):

    def test(self):

        # Check the stored event class doesn't function before the connection is setup.
        with self.assertRaises(DatastoreConnectionError):
            self.list_records()
        with self.assertRaises(DatastoreConnectionError):
            self.create_record()

        # Setup the connection.
        self.datastore.setup_connection()

        # Check it doesn't matter if setup_connection() is called twice.
        self.datastore.setup_connection()

        # Check the stored event class doesn't function before the tables are setup.
        with self.assertRaises(DatastoreTableError):
            self.list_records()
        with self.assertRaises(DatastoreTableError):
            self.create_record()

        # Setup the tables.
        self.datastore.setup_tables()

        # Check it doesn't matter if setup_tables() is called twice.
        self.datastore.setup_tables()

        # Check the stored event class does function after the tables have been setup.
        self.assertEqual(len(self.list_records()), 0)
        self.create_record()
        self.assertEqual(len(self.list_records()), 1)

        # Drop the tables.
        self.datastore.drop_tables()

        # Check the stored event class doesn't function after the tables have been dropped.
        with self.assertRaises(DatastoreTableError):
            self.list_records()
        with self.assertRaises(DatastoreTableError):
            self.create_record()

        # Drop the tables.
        self.datastore.drop_connection()

        # Check the stored event class doesn't function after the connection has been dropped.
        with self.assertRaises(DatastoreConnectionError):
            self.list_records()
        with self.assertRaises(DatastoreConnectionError):
            self.create_record()

    @abstractmethod
    def list_records(self):
        return []

    @abstractmethod
    def create_record(self):
        return None

    def tearDown(self):
        # Try to remove any tables.
        self.datastore.setup_connection()
        self.datastore.drop_tables()
        self.datastore.drop_connection()
        super(DatastoreTestCase, self).tearDown()




from math import floor
from time import time
from uuid import uuid4

import six

from eventsourcing.domain.model.timebucketedlog import start_new_timebucketedlog
from eventsourcing.example.domainmodel import Example, register_new_example
from eventsourcing.infrastructure.activerecord import AbstractActiveRecordStrategy
from eventsourcing.infrastructure.eventstore import EventStore
from eventsourcing.infrastructure.iterators import SequencedItemIterator
from eventsourcing.infrastructure.timebucketedlog_reader import TimebucketedlogReader, get_timebucketedlog_reader
from eventsourcing.tests.base import notquick
from eventsourcing.tests.core_tests.test_utils import utc_now
from eventsourcing.tests.example_application_tests.base import WithExampleApplication
from eventsourcing.tests.example_application_tests.test_example_application_with_encryption import \
    WithEncryption
from eventsourcing.tests.sequenced_item_tests.test_cassandra_active_record_strategy import \
    WithCassandraActiveRecordStrategies
from eventsourcing.tests.sequenced_item_tests.test_sqlalchemy_active_record_strategy import \
    WithSQLAlchemyActiveRecordStrategies


@notquick()
class PerformanceTestCase(WithExampleApplication):
    def test(self):
        """
        Reports on the performance of Example entity and repo.

        NB: This test doesn't actually assert anything, so it isn't really a test.
        """

        with self.construct_application() as app:

            # Initialise dict of entities.
            self.entities = {}

            report_name = type(self).__name__[4:]
            print("\n\n{} report:\n".format(report_name))

            repetitions = 1  # 10

            # NB: Use range(1, 5) to test whether we can get more than 10000 items from Cassandra.
            for i in six.moves.range(0, 5):
                # Setup a number of entities, with different lengths of event history.
                payload = 3
                # payload = str([uuid4().hex for _ in six.moves.range(100000)])
                example = register_new_example(a=1, b=payload)
                self.entities[i] = example

                # Beat a number of times.
                start_beating = utc_now()
                num_beats = int(floor(10 ** i))
                for _ in six.moves.range(num_beats):
                    example.beat_heart()
                time_beating = utc_now() - start_beating
                print("Time to beat {} times: {:.2f}s ({:.0f} beats/s, {:.6f}s each)"
                      "".format(num_beats, time_beating, num_beats / time_beating, time_beating / num_beats))

                # Get the last n events from the repo.
                def last_n(n):
                    n = min(n, num_beats + 1)
                    assert isinstance(app.example_repo.event_player.event_store, EventStore)
                    ars = app.example_repo.event_player.event_store.active_record_strategy
                    assert isinstance(ars, AbstractActiveRecordStrategy)

                    start_last_n = utc_now()
                    last_n_stored_events = []
                    for _ in six.moves.range(repetitions):
                        iterator = SequencedItemIterator(
                            active_record_strategy=ars,
                            sequence_id=example.id,
                            limit=n,
                            is_ascending=False,
                        )
                        last_n_stored_events = list(iterator)
                    time_last_n = (utc_now() - start_last_n) / repetitions

                    num_retrieved_events = len(last_n_stored_events)
                    events_per_second = num_retrieved_events / time_last_n
                    print(("Time to get last {:>" + str(i + 1) + "} events after {} events: {:.6f}s ({:.0f} events/s)"
                                                                 "").format(n, num_beats + 1, time_last_n,
                                                                            events_per_second))

                for j in range(0, i + 1):
                    last_n(10 ** j)

                # Get the entity by replaying all events (which it must since there isn't a snapshot).
                start_replay = utc_now()
                for _ in six.moves.range(repetitions):
                    example = app.example_repo[example.id]
                    assert isinstance(example, Example)
                    heartbeats = example.count_heartbeats()
                    assert heartbeats == num_beats, (heartbeats, num_beats)

                time_replaying = (utc_now() - start_replay) / repetitions
                print("Time to replay {} beats: {:.2f}s ({:.0f} beats/s, {:.6f}s each)"
                      "".format(num_beats, time_replaying, num_beats / time_replaying, time_replaying / num_beats))

                # Take snapshot, and beat heart a few more times.
                app.example_repo.event_player.take_snapshot(example.id, lte=time())

                extra_beats = 4
                for _ in six.moves.range(extra_beats):
                    example.beat_heart()
                num_beats += extra_beats

                # Get the entity using snapshot and replaying events since the snapshot.
                start_replay = utc_now()
                for _ in six.moves.range(repetitions):
                    example = app.example_repo[example.id]
                time_replaying = (utc_now() - start_replay) / repetitions

                events_per_second = (extra_beats + 1) / time_replaying  # +1 for the snapshot event
                beats_per_second = num_beats / time_replaying
                print("Time to replay snapshot with {} extra beats: {:.6f}s ({:.0f} events/s, {:.0f} beats/s)"
                      "".format(extra_beats, time_replaying, events_per_second, beats_per_second))

                print("")

    def test_log_performance(self):

        with self.construct_application() as app:
            example_id = uuid4()
            log = start_new_timebucketedlog(example_id, bucket_size='year')
            log_reader = get_timebucketedlog_reader(log, app.timestamp_entity_event_store)

            # Write a load of messages.
            start_write = utc_now()
            number_of_messages = 111
            events = []
            for i in range(number_of_messages):
                event = log.append_message('Logger message number {}'.format(i))
                events.append(event)
            time_to_write = (utc_now() - start_write)
            print("Time to log {} messages: {:.2f}s ({:.0f} messages/s, {:.6f}s each)"
                  "".format(number_of_messages, time_to_write, number_of_messages / time_to_write,
                            time_to_write / number_of_messages))

            # Read pages of messages in descending order.
            # - get a limited number until a time, then use the earliest in that list as the position
            position = events[-1].timestamp

            page_size = 10

            # Page back through the log in reverse chronological order.
            previous_position = None
            next_position = None
            count_pages = 0
            total_time_to_read = 0
            total_num_reads = 0
            while True:
                start_read = utc_now()
                page_of_events, next_position = self.get_message_logged_events_and_next_position(log_reader, position,
                                                                                                 page_size)
                time_to_read = (utc_now() - start_read)
                total_time_to_read += time_to_read
                total_num_reads += 1
                count_pages += 1
                if next_position is None:
                    break
                else:
                    previous_position, position = position, next_position

            # Check we got to the end of the line.
            self.assertEqual(count_pages, 11)
            self.assertIsNone(next_position)
            self.assertTrue(previous_position)

            # Page forward through the log in chronological order.
            count_pages = 0
            position = None
            while True:
                start_read = utc_now()
                page_of_events, next_position = self.get_message_logged_events_and_next_position(log_reader, position,
                                                                                                 page_size,
                                                                                                 is_ascending=True)
                time_to_read = (utc_now() - start_read)
                total_time_to_read += time_to_read
                total_num_reads += 1
                count_pages += 1
                if next_position is None:
                    break
                else:
                    position = next_position

            self.assertEqual(count_pages, 11)
            self.assertIsNone(next_position)
            self.assertTrue(previous_position)

            reads_per_second = total_num_reads / total_time_to_read
            messages_per_second = reads_per_second * number_of_messages
            print("Time to read {} pages of logged messages: {:.6f}s ({:.0f} pages/s, {:.0f} messages/s))"
                  "".format(total_num_reads, total_time_to_read, reads_per_second, messages_per_second))

    def get_message_logged_events_and_next_position(self, log_reader, position, page_size, is_ascending=False):
        assert isinstance(log_reader, TimebucketedlogReader), type(log_reader)
        assert isinstance(page_size, six.integer_types), type(page_size)
        assert isinstance(is_ascending, bool)
        if is_ascending:
            gt = position
            lt = None
        else:
            lt = position
            gt = None

        events = log_reader.get_events(gt=gt, lt=lt, limit=page_size + 1, is_ascending=is_ascending)
        events = list(events)
        if len(events) == page_size + 1:
            next_position = events.pop().timestamp
        else:
            next_position = None
        return events, next_position


@notquick()
class TestCassandraPerformance(WithCassandraActiveRecordStrategies, PerformanceTestCase):
    pass


@notquick()
class TestEncryptionPerformance(WithEncryption, TestCassandraPerformance):
    pass


@notquick()
class TestSQLAlchemyPerformance(WithSQLAlchemyActiveRecordStrategies, PerformanceTestCase):
    pass



from tempfile import NamedTemporaryFile

from sqlalchemy.exc import OperationalError

from eventsourcing.infrastructure.datastore import DatastoreTableError
from eventsourcing.infrastructure.sqlalchemy.datastore import DEFAULT_SQLALCHEMY_DB_URI, SQLAlchemyDatastore, \
    SQLAlchemySettings, Base
from eventsourcing.infrastructure.sqlalchemy.activerecords import SqlIntegerSequencedItem, SqlTimestampSequencedItem
from eventsourcing.tests.datastore_tests.base import AbstractDatastoreTestCase, DatastoreTestCase


class SQLAlchemyDatastoreTestCase(AbstractDatastoreTestCase):
    use_named_temporary_file = False

    def construct_datastore(self):
        if self.use_named_temporary_file:
            self.temp_file = NamedTemporaryFile('a', delete=True)
            uri = 'sqlite:///' + self.temp_file.name
        else:
            uri = DEFAULT_SQLALCHEMY_DB_URI
        return SQLAlchemyDatastore(
            base=Base,
            settings=SQLAlchemySettings(uri=uri),
            tables=(SqlIntegerSequencedItem, SqlTimestampSequencedItem),
        )


class TestSQLAlchemyDatastore(SQLAlchemyDatastoreTestCase, DatastoreTestCase):
    def list_records(self):
        try:
            query = self.datastore.db_session.query(SqlIntegerSequencedItem)
            return list(query)
        except OperationalError as e:
            self.datastore.db_session.rollback()
            raise DatastoreTableError(e)

    def create_record(self):
        try:
            record = SqlIntegerSequencedItem()
            self.datastore.db_session.add(record)
            self.datastore.db_session.commit()
        except OperationalError as e:
            self.datastore.db_session.rollback()
            raise DatastoreTableError(e)
        return record



# from singledispatch import singledispatch
#
# from eventsourcing.domain.model.entity import AbstractEntityRepository, Aggregate, Created, entity_mutator
# from eventsourcing.domain.model.events import publish
# from eventsourcing.exceptions import RepositoryKeyError
#
#
# class NotificationLog(Aggregate):
#     class Started(Created):
#         pass
#
#     def __init__(self, name, bucket_size=None, sequence_size=None, **kwargs):
#         super(NotificationLog, self).__init__(**kwargs)
#         self._name = name
#         self._sequence_size = sequence_size
#         self._bucket_size = bucket_size
#
#     @property
#     def name(self):
#         return self._name
#
#     @property
#     def sequence_size(self):
#         return self._sequence_size
#
#     @property
#     def bucket_size(self):
#         return self._bucket_size
#
#     @staticmethod
#     def _mutator(event, initial):
#         return notification_log_mutator(event, initial)
#
#
# @singledispatch
# def notification_log_mutator(event, initial):
#     return entity_mutator(event, initial)
#
#
# class NotificationLogRepository(AbstractEntityRepository):
#     def get_or_create(self, log_name, timebucket_size=None, sequence_size=None):
#         """
#         Gets or creates a log.
#
#         :rtype: NotificationLog
#         """
#         try:
#             return self[log_name]
#         except RepositoryKeyError:
#             return start_notification_log(
#                 log_name=log_name,
#                 timebucket_size=timebucket_size,
#                 sequence_size=sequence_size,
#             )
#
#
# def start_notification_log(log_name, timebucket_size=None, sequence_size=None):
#     event = NotificationLog.Started(
#         entity_id=log_name,
#         name=log_name,
#         bucket_size=timebucket_size,
#         sequence_size=sequence_size,
#     )
#     entity = NotificationLog.mutate(event=event)
#     publish(event)
#     return entity



from eventsourcing.domain.model.events import subscribe


def subscribe_to(event_class):
    """ Decorator for making a custom event handler function subscribe to a certain event type
    Args:
        event_class: DomainEvent class or its child classes that the handler function should subscribe to

    Example usage:
        # this example shows a custom handler that reacts to Todo.Created
        # event and saves a projection of a Todo model object
        @subscribe_to(Todo.Created)
        def new_todo_projection(event):
            todo = TodoProjection(id=event.entity_id, title=event.title)
            todo.save()
    """

    def event_type_predicate(event):
        return isinstance(event, event_class)

    def wrap(handler_func):
        subscribe(handler_func, event_type_predicate)
        return handler_func

    return wrap



from eventsourcing.example.domainmodel import Example, AbstractExampleRepository
from eventsourcing.infrastructure.eventsourcedrepository import EventSourcedRepository


class ExampleRepository(EventSourcedRepository, AbstractExampleRepository):
    """
    Event sourced repository for the Example domain model entity.
    """
    __page_size__ = 1000
    mutator = Example.mutate




from time import time

import six
from six import with_metaclass

from eventsourcing.domain.model.events import QualnameABCMeta
from eventsourcing.domain.model.timebucketedlog import MessageLogged, Timebucketedlog, make_timebucket_id, \
    next_bucket_starts, previous_bucket_starts
from eventsourcing.infrastructure.eventstore import AbstractEventStore


def get_timebucketedlog_reader(log, event_store):
    """
    :rtype: TimebucketedlogReader
    """
    return TimebucketedlogReader(log=log, event_store=event_store)


class TimebucketedlogReader(with_metaclass(QualnameABCMeta)):
    def __init__(self, log, event_store, page_size=50):
        assert isinstance(log, Timebucketedlog)
        self.log = log
        assert isinstance(event_store, AbstractEventStore), event_store
        self.event_store = event_store
        assert isinstance(page_size, six.integer_types)
        self.page_size = page_size
        self.position = None

    def get_messages(self, gt=None, gte=None, lt=None, lte=None, limit=None, is_ascending=False, page_size=None):
        events = self.get_events(gt=gt, gte=gte, lt=lt, lte=lte, limit=limit, is_ascending=is_ascending,
                                 page_size=page_size)
        for event in events:
            if isinstance(event, MessageLogged):
                self.position = event.timestamp
                yield event.message

    def get_events(self, gt=None, gte=None, lt=None, lte=None, limit=None, is_ascending=False, page_size=None):
        assert limit is None or limit > 0

        # Identify the first time bucket.
        now = time()
        started_on = self.log.started_on
        absolute_latest = min(now, lt or now, lte or now)
        absolute_earlyist = max(started_on, gt or 0, gte or 0)
        if is_ascending:
            position = absolute_earlyist
        else:
            position = absolute_latest

        # Start counting events.
        count_events = 0

        while True:
            bucket_id = make_timebucket_id(self.log.name, position, self.log.bucket_size)
            for message_logged_event in self.event_store.get_domain_events(
                entity_id=bucket_id,
                gt=gt,
                gte=gte,
                lt=lt,
                lte=lte,
                limit=limit,
                is_ascending=is_ascending,
                page_size=page_size,
            ):
                yield message_logged_event

                if limit is not None:
                    count_events += 1
                    if count_events >= limit:
                        raise StopIteration

            # See if there's another bucket.
            if is_ascending:
                next_timestamp = next_bucket_starts(position, self.log.bucket_size)
                if next_timestamp > absolute_latest:
                    raise StopIteration
                else:
                    position = next_timestamp
            else:
                if position < absolute_earlyist:
                    raise StopIteration
                else:
                    position = previous_bucket_starts(position, self.log.bucket_size)



from eventsourcing.infrastructure.cassandra.activerecords import CassandraActiveRecordStrategy, \
    CqlIntegerSequencedItem, CqlTimestampSequencedItem
from eventsourcing.infrastructure.sequenceditem import SequencedItem
from eventsourcing.tests.datastore_tests.test_cassandra import CassandraDatastoreTestCase
from eventsourcing.tests.sequenced_item_tests.base import IntegerSequencedItemTestCase, \
    SimpleSequencedItemteratorTestCase, ThreadedSequencedItemIteratorTestCase, TimestampSequencedItemTestCase, \
    WithActiveRecordStrategies


def construct_integer_sequenced_active_record_strategy():
    return CassandraActiveRecordStrategy(
        active_record_class=CqlIntegerSequencedItem,
        sequenced_item_class=SequencedItem,
    )


def construct_timestamp_sequenced_active_record_strategy():
    return CassandraActiveRecordStrategy(
        active_record_class=CqlTimestampSequencedItem,
        sequenced_item_class=SequencedItem,
    )


class TestCassandraActiveRecordStrategyWithIntegerSequences(CassandraDatastoreTestCase,
                                                            IntegerSequencedItemTestCase):
    def construct_active_record_strategy(self):
        return construct_integer_sequenced_active_record_strategy()


class TestCassandraActiveRecordStrategyWithTimestampSequences(CassandraDatastoreTestCase,
                                                              TimestampSequencedItemTestCase):
    def construct_active_record_strategy(self):
        return construct_timestamp_sequenced_active_record_strategy()


class WithCassandraActiveRecordStrategies(CassandraDatastoreTestCase, WithActiveRecordStrategies):
    def construct_integer_sequenced_active_record_strategy(self):
        return construct_integer_sequenced_active_record_strategy()

    def construct_timestamp_sequenced_active_record_strategy(self):
        return construct_timestamp_sequenced_active_record_strategy()


class TestSimpleSequencedItemIteratorWithCassandra(WithCassandraActiveRecordStrategies,
                                                   SimpleSequencedItemteratorTestCase):
    pass


class TestThreadedSequencedItemIteratorWithCassandra(WithCassandraActiveRecordStrategies,
                                                     ThreadedSequencedItemIteratorTestCase):
    pass



class EventSourcingError(Exception):
    """Base eventsourcing exception."""


class TopicResolutionError(EventSourcingError):
    """Raised when unable to resolve a topic to a Python class."""


class EntityVersionNotFound(EventSourcingError):
    """Raise when accessing an entity version that does not exist."""


class ConcurrencyError(EventSourcingError):
    """Raised when appending events at the wrong version to a versioned stream."""


class ConsistencyError(EventSourcingError):
    """Raised when applying an event stream to a versioned entity."""


class ProgrammingError(EventSourcingError):
    """Raised when programming errors are encountered."""


class RepositoryKeyError(KeyError, EventSourcingError):
    """Raised when access an entity that does not exist."""


class SequenceFullError(EventSourcingError):
    "Raised when attempting to append an item to a sequence that is already at its maximum size."


class DatasourceSettingsError(EventSourcingError):
    "Raised when an error is detected in settings for a datasource."


class SequencedItemError(EventSourcingError):
    "Raised when an integer sequence error occurs e.g. trying to save a version that already exists."


class TimeSequenceError(EventSourcingError):
    "Raised when a time sequence error occurs e.g. trying to save a timestamp that already exists."



from unittest.case import TestCase
from uuid import uuid1

import datetime

from eventsourcing.utils.time import timestamp_from_uuid, utc_timezone


class TestUtils(TestCase):

    def test_timestamp_from_uuid(self):
        until = utc_now()
        uuid = uuid1()
        after = utc_now()
        uuid_timestamp = timestamp_from_uuid(uuid)
        self.assertLess(until, uuid_timestamp)
        self.assertGreater(after, uuid_timestamp)

        # Check it works with hex strings, as well as UUID objects.
        self.assertEqual(timestamp_from_uuid(uuid.hex), timestamp_from_uuid(uuid))


def utc_now():
    now_datetime = datetime.datetime.now(utc_timezone)
    try:
        now_timestamp = now_datetime.timestamp()
    except AttributeError:
        now_timestamp = (now_datetime - datetime.datetime(1970, 1, 1, tzinfo=utc_timezone)).total_seconds()
    return now_timestamp



from abc import ABCMeta, abstractproperty

import six

from eventsourcing.domain.model.events import TimestampedEntityEvent


class AbstractSnapshop(six.with_metaclass(ABCMeta)):

    @abstractproperty
    def topic(self):
        """
        Path to the class of the snapshotted entity.
        """

    @abstractproperty
    def state(self):
        """
        State of the snapshotted entity.
        """

    @abstractproperty
    def timestamp(self):
        """
        Timestamp of the snapshot.
        """


class Snapshot(TimestampedEntityEvent, AbstractSnapshop):

    def __init__(self, entity_id, timestamp, topic, state):
        super(Snapshot, self).__init__(
            entity_id=entity_id,
            timestamp=timestamp,
            topic=topic,
            state=state,
        )

    @property
    def timestamp(self):
        return self.__dict__['timestamp']

    @property
    def topic(self):
        """Path to the class.
        """
        return self.__dict__['topic']

    @property
    def state(self):
        """
        Snapshotted state of the entity.
        """
        return self.__dict__['state']



from functools import reduce

from copy import deepcopy

from eventsourcing.domain.model.entity import TimestampedVersionedEntity
from eventsourcing.domain.model.snapshot import AbstractSnapshop
from eventsourcing.infrastructure.eventstore import AbstractEventStore
from eventsourcing.infrastructure.snapshotting import AbstractSnapshotStrategy, entity_from_snapshot


# def clone_object(initial_state):
#     initial_state_copy = object.__new__(type(initial_state))
#     initial_state_copy.__dict__.update(deepcopy(initial_state.__dict__))
#     return initial_state_copy


class EventPlayer(object):
    """
    Reconstitutes domain entities from domain events
    retrieved from the event store, optionally with snapshots.
    """

    def __init__(self, event_store, mutator, page_size=None, is_short=False, snapshot_strategy=None):
        assert isinstance(event_store, AbstractEventStore), event_store
        if snapshot_strategy is not None:
            assert isinstance(snapshot_strategy, AbstractSnapshotStrategy), snapshot_strategy
        self.event_store = event_store
        self.mutator = mutator
        self.page_size = page_size
        self.is_short = is_short
        self.snapshot_strategy = snapshot_strategy

    def replay_entity(self, entity_id, gt=None, gte=None, lt=None, lte=None, limit=None, initial_state=None,
                      query_descending=False):
        """
        Reconstitutes requested domain entity from domain events found in event store.
        """
        # Decide if query is in ascending order.
        #  - A "speed up" for when events are stored in descending order (e.g.
        #  in Cassandra) and it is faster to get them in that order.
        #  - This isn't useful when 'until' or 'after' or 'limit' are set,
        #    because the inclusiveness or exclusiveness of until and after
        #    and the end of the stream that is truncated by limit both depend on
        #    the direction of the query. Also paging backwards isn't useful, because
        #    all the events are needed eventually, so it would probably slow things
        #    down. Paging is intended to support replaying longer event streams, and
        #    only makes sense to work in ascending order.
        if self.is_short and gt is None and gte is None and lt is None and lte is None and self.page_size is None:
            is_ascending = False
        else:
            is_ascending = not query_descending

        # Get the domain events that are to be replayed.
        domain_events = self.get_domain_events(entity_id,
                                               gt=gt,
                                               gte=gte,
                                               lt=lt,
                                               lte=lte,
                                               limit=limit,
                                               is_ascending=is_ascending)

        # The events will be replayed in ascending order.
        if not is_ascending:
            domain_events = reversed(list(domain_events))

        # Replay the domain events, starting with the initial state.
        return self.replay_events(initial_state, domain_events)

    def replay_events(self, initial_state, domain_events):
        """
        Mutates initial state using the sequence of domain events.
        """
        return reduce(self.mutator, domain_events, initial_state)

    def get_domain_events(self, entity_id, gt=None, gte=None, lt=None, lte=None, limit=None, is_ascending=True):
        """
        Returns domain events for given entity ID.
        """
        # Get entity's domain events from the event store.
        domain_events = self.event_store.get_domain_events(
            entity_id=entity_id,
            gt=gt,
            gte=gte,
            lt=lt,
            lte=lte,
            limit=limit,
            page_size=self.page_size,
            is_ascending=is_ascending,
        )
        return domain_events

    def take_snapshot(self, entity_id, lt=None, lte=None):
        """
        Takes a snapshot of the entity as it existed after the most recent
        event, optionally less than or less than or equal to a particular position.
        """
        assert isinstance(self.snapshot_strategy, AbstractSnapshotStrategy)

        # Get the last event (optionally until a particular time).
        last_event = self.get_most_recent_event(entity_id, lt=lt, lte=lte)

        # If there aren't any events, there can't be a snapshot, so return None.
        if last_event is None:
            return None

        # If there is something to snapshot, then look for
        # the last snapshot before the last event.
        last_snapshot = self.get_snapshot(entity_id, lte=last_event.timestamp)

        if last_snapshot:
            assert isinstance(last_snapshot, AbstractSnapshop), type(last_snapshot)
            if last_snapshot.timestamp < last_event.timestamp:
                # There must be events after the snapshot, so get events after
                # the last event that was applied to the entity, and obtain the
                # initial entity state so those event can be applied to it.
                initial_state = entity_from_snapshot(last_snapshot)
                gte = initial_state._version
            else:
                # There's nothing to do.
                return last_snapshot
        else:
            # If there isn't a snapshot, start from scratch.
            initial_state = None
            gte = None

        # Get entity in the state after this event was applied.
        entity = self.replay_entity(entity_id, gte=gte, lt=lt, lte=lte, initial_state=initial_state)

        # Take a snapshot of the entity.
        return self.snapshot_strategy.take_snapshot(entity, timestamp=entity.last_modified_on)

    def get_snapshot(self, entity_id, lt=None, lte=None):
        """
        Returns a snapshot for given entity ID, according to the snapshot strategy.
        """
        if self.snapshot_strategy:
            return self.snapshot_strategy.get_snapshot(entity_id, lt=lt, lte=lte)

    def get_most_recent_event(self, entity_id, lt=None, lte=None):
        """
        Returns the most recent event for the given entity ID.
        """
        return self.event_store.get_most_recent_event(entity_id, lt=lt, lte=lte)

    # def fastforward(self, stale_entity, lt=None, lte=None):
    #     assert isinstance(stale_entity, TimestampedVersionedEntity)
    #
    #     # Replay the events since the entity version.
    #     fresh_entity = self.replay_entity(
    #         entity_id=stale_entity.id,
    #         gt=stale_entity.version,
    #         lt=lt,
    #         lte=lte,
    #         initial_state=stale_entity,
    #     )
    #
    #     # Return the fresh instance.
    #     return fresh_entity



from __future__ import unicode_literals

import json
from abc import ABCMeta, abstractmethod

import six

from eventsourcing.domain.model.events import resolve_domain_topic, topic_from_domain_class
from eventsourcing.domain.services.cipher import AbstractCipher
from eventsourcing.infrastructure.sequenceditem import SequencedItemFieldNames
from eventsourcing.infrastructure.transcoding import ObjectJSONDecoder, ObjectJSONEncoder


class AbstractSequencedItemMapper(six.with_metaclass(ABCMeta)):
    @abstractmethod
    def to_sequenced_item(self, domain_event):
        """Serializes a domain event."""

    @abstractmethod
    def from_sequenced_item(self, serialized_event):
        """Deserializes domain events."""


class SequencedItemMapper(AbstractSequencedItemMapper):
    """
    Uses JSON to transcode domain events.
    """

    def __init__(self, sequenced_item_class, event_sequence_id_attr, event_position_attr,
                 json_encoder_class=ObjectJSONEncoder, json_decoder_class=ObjectJSONDecoder,
                 always_encrypt=False, cipher=None):
        self.sequenced_item_class = sequenced_item_class
        self.event_sequence_id_attr = event_sequence_id_attr
        self.event_position_attr = event_position_attr
        self.json_encoder_class = json_encoder_class
        self.json_decoder_class = json_decoder_class
        self.cipher = cipher
        self.always_encrypt = always_encrypt
        self.field_names = SequencedItemFieldNames(self.sequenced_item_class)

    def to_sequenced_item(self, domain_event):
        """
        Constructs a sequenced item from a domain event.
        """
        item_args = self.construct_item_args(domain_event)
        return self.construct_sequenced_item(item_args)

    def construct_item_args(self, domain_event):
        # Construct attributes of a sequenced item from the domain event.

        # Identify the sequence ID.
        sequence_id = getattr(domain_event, self.event_sequence_id_attr)

        # Identify the position in the sequence.
        position = getattr(domain_event, self.event_position_attr)
        topic = self.topic_from_domain_class(domain_event.__class__)

        # Serialise event attributes to JSON, optionally encrypted.
        is_encrypted = self.is_encrypted(domain_event.__class__)
        event_data = self.serialize_event_attrs(domain_event.__dict__, is_encrypted=is_encrypted)

        return (sequence_id, position, topic, event_data)

    def construct_sequenced_item(self, item_args):
        return self.sequenced_item_class(*item_args)

    def from_sequenced_item(self, sequenced_item):
        """
        Reconstructs domain event from stored event topic and
        event attrs. Used in the event store when getting domain events.
        """
        assert isinstance(sequenced_item, self.sequenced_item_class), (self.sequenced_item_class, type(sequenced_item))

        # Get the domain event class from the topic.
        domain_event_class = self.class_from_topic(getattr(sequenced_item, self.field_names.topic))

        # Deserialize, optionally with decryption.
        is_encrypted = self.is_encrypted(domain_event_class)
        event_attrs = self.deserialize_event_attrs(getattr(sequenced_item, self.field_names.data), is_encrypted)

        # Reconstruct the domain event object.
        return self.reconstruct_object(domain_event_class, event_attrs)

    def reconstruct_object(self, obj_class, obj_state):
        obj = object.__new__(obj_class)
        obj.__dict__.update(obj_state)
        return obj

    def topic_from_domain_class(self, domain_class):
        return topic_from_domain_class(domain_class)

    def class_from_topic(self, topic):
        return resolve_domain_topic(topic)

    def serialize_event_attrs(self, event_attrs, is_encrypted=False):
        event_data = json.dumps(
            event_attrs,
            separators=(',', ':'),
            sort_keys=True,
            cls=self.json_encoder_class,
        )
        # Encrypt (optional).
        if is_encrypted:
            assert isinstance(self.cipher, AbstractCipher)
            event_data = self.cipher.encrypt(event_data)

        return event_data

    def deserialize_event_attrs(self, event_attrs, is_encrypted):
        """
        Deserialize event attributes from JSON, optionally with decryption.
        """
        if is_encrypted:
            assert isinstance(self.cipher, AbstractCipher), self.cipher
            event_attrs = self.cipher.decrypt(event_attrs)
        return json.loads(event_attrs, cls=self.json_decoder_class)

    def is_encrypted(self, domain_event_class):
        return self.always_encrypt or getattr(domain_event_class, '__always_encrypt__', False)


def deserialize_domain_entity(entity_topic, entity_attrs):
    """
    Return a new domain entity object from a given topic (a string) and attributes (a dict).
    """

    # Get the domain entity class from the entity topic.
    domain_class = resolve_domain_topic(entity_topic)

    # Instantiate the domain entity class.
    entity = object.__new__(domain_class)

    # Set the attributes.
    entity.__dict__.update(entity_attrs)

    # Return a new domain entity object.
    return entity



from eventsourcing.domain.model.sequence import SequenceRepository, Sequence
from eventsourcing.infrastructure.eventsourcedrepository import EventSourcedRepository


class SequenceRepo(EventSourcedRepository, SequenceRepository):
    mutator = Sequence.mutate

    def get_entity(self, entity_id, until=None):
        """
        Replays entity using only the 'Started' event.
        """
        return self.event_player.replay_entity(entity_id, limit=1)



import six

from eventsourcing.domain.model.events import publish
from eventsourcing.domain.model.sequence import Sequence
from eventsourcing.exceptions import SequenceFullError
from eventsourcing.infrastructure.eventplayer import EventPlayer


def append_item_to_sequence(name, item, event_player, max_size=None):
    assert isinstance(event_player, EventPlayer)
    last_event = event_player.event_store.get_most_recent_event(name)
    next_version = last_event.entity_version + 1
    if max_size and max_size < next_version:
        raise SequenceFullError
    event = Sequence.Appended(
        entity_id=name,
        entity_version=next_version,
        item=item,
    )
    publish(event)


class SequenceReader(object):
    def __init__(self, sequence, event_player):
        assert isinstance(sequence, Sequence), sequence
        assert isinstance(event_player, EventPlayer), event_player
        self.sequence = sequence
        self.event_player = event_player

    def __getitem__(self, item):
        assert isinstance(item, (six.integer_types, slice))
        sequence_len = None
        if isinstance(item, six.integer_types):
            if item < 0:
                if sequence_len is None:
                    sequence_len = len(self)
                index = sequence_len + item
                if index < 0:
                    raise IndexError("Sequence index out of range: {}".format(item))
            else:
                index = item
            event = self.event_player.event_store.get_domain_event(
                entity_id=self.sequence.id,
                eq=index + 1,
            )
            return event.item
        elif isinstance(item, slice):
            assert item.step == None, "Slice step must be 1: {}".format(str(item.step))
            if item.start is None:
                start_index = 0
            elif item.start < 0:
                if sequence_len is None:
                    sequence_len = len(self)
                start_index = max(sequence_len + item.start, 0)
            else:
                start_index = item.start

            if not isinstance(item.stop, six.integer_types):
                limit = None
            elif item.stop < 0:
                if sequence_len is None:
                    sequence_len = len(self)
                limit = sequence_len + item.stop - start_index
            else:
                limit = item.stop - start_index

            if limit is not None and limit <= 0:
                return []

            events = self.event_player.event_store.get_domain_events(
                entity_id=self.sequence.id,
                gte=start_index + 1,
                limit=limit
            )
            items = [e.item for e in events]
            return items

    def __len__(self):
        events = self.event_player.get_domain_events(self.sequence.id, is_ascending=False, limit=1)
        return events[0].entity_version



# import time
#
# from eventsourcing.exceptions import ConcurrencyError
#
#
# def retry_after_concurrency_error(operation_to_retry, max_attempts=10, wait=0.02, args=()):
#     tries = 0
#     assert tries <= max_attempts
#     while True:
#         try:
#             operation_to_retry(*args)
#             break
#         except ConcurrencyError as error:
#             tries += 1
#             if tries < max_attempts:
#                 time.sleep(wait * (1 + tries))
#                 continue
#             else:
#                 raise error



# from threading import Thread
# from unittest.case import TestCase
#
# from eventsourcing.infrastructure.event_sourced_repos.notificationlog_repo import NotificationLogRepo
# from eventsourcing.infrastructure.event_sourced_repos.sequence import SequenceRepo
# from eventsourcing.infrastructure.notification_log import append_item_to_notification_log
# from eventsourcing.interface.archived_logs import ArchivedLogReader, ArchivedLogRepo, RemoteArchivedLogRepo, \
#     serialize_archived_log, deserialise_archived_log
# from eventsourcing.tests.sequenced_item_tests.test_cassandra_active_record_strategy import \
#     WithCassandraActiveRecordStrategies
# from eventsourcing.tests.sequenced_item_tests.test_sqlalchemy_active_record_strategy import \
#     WithSQLAlchemyActiveRecordStrategies
# from eventsourcing.tests.sequenced_item_tests.base import WithPersistencePolicy
#
#
# class NotificationLogContext(object):
#
#     def __init__(self, event_store):
#         super(NotificationLogContext, self).__init__()
#         self.event_store = event_store
#         self.sequence_repo = SequenceRepo(event_store)
#         self.notification_log_repo = NotificationLogRepo(event_store)
#
#     def get_or_create_notification_log(self, log_name, sequence_size):
#         return self.notification_log_repo.get_or_create(
#             log_name=log_name,
#             sequence_size=sequence_size,
#         )
#
#     def append_notification(self, notification_log, item):
#         append_item_to_notification_log(
#             notification_log=notification_log,
#             item=item,
#             sequence_repo=self.sequence_repo,
#             log_repo=self.log_repo,
#             event_store=self.sequence_repo.event_store,
#         )
#
#     def create_archived_log_repo(self, notification_log, doc_size):
#         return ArchivedLogRepo(
#             notification_log=notification_log,
#             sequence_repo=self.sequence_repo,
#             log_repo=self.log_repo,
#             event_store=self.event_store,
#             doc_size=doc_size,
#         )
#
#
# class ArchivedLogTestCase(WithPersistencePolicy):
#
#     def setUp(self):
#         super(ArchivedLogTestCase, self).setUp()
#         self.app = NotificationLogContext(self.versioned_entity_event_store)
#
#     def test_archived_log_repo(self):
#         notification_log = self.app.get_or_create_notification_log('log1', sequence_size=10)
#
#         # Create an archived log repo.
#         doc_size = 5
#         repo = self.app.create_archived_log_repo(notification_log, doc_size)
#
#         # Check the archived logs.
#         archived_log = repo['current']
#         self.assertEqual(archived_log.id, '1,5')
#         self.assertEqual(len(archived_log.items), 0)
#         self.assertIsNone(archived_log.previous_id)
#         self.assertIsNone(archived_log.previous_id)
#
#         # Append items to notification log.
#         self.append_notifications(notification_log, 13)
#
#         # Check the archived logs.
#         self.assert_archived_log(repo, 'current', '11,15', 3, '6,10', None)
#         self.assert_archived_log(repo, '1,5', '1,5', doc_size, None, '6,10')
#         self.assert_archived_log(repo, '6,10', '6,10', doc_size, '1,5', '11,15')
#         self.assert_archived_log(repo, '11,15', '11,15', 3, '6,10', None)
#         self.assert_archived_log(repo, '16,20', '16,20', 0, '11,15', None)
#         self.assert_archived_log(repo, '21,25', '21,25', 0, '16,20', None)
#
#         # Add some more items.
#         self.append_notifications(notification_log, 13, 24)
#
#         # Check the archived logs have been extended.
#         self.assertEqual(len(repo['11,15'].items), doc_size)
#         self.assertEqual(len(repo['16,20'].items), doc_size)
#         self.assertEqual(len(repo['21,25'].items), 4)
#         self.assertEqual(len(repo['26,30'].items), 0)
#
#         # Check sequence size must be divisible by doc size.
#         with self.assertRaises(ValueError):
#             self.app.create_archived_log_repo(notification_log, doc_size=6)
#
#         # Check the doc ID must match the doc size.
#         repo = self.app.create_archived_log_repo(notification_log, doc_size)
#         with self.assertRaises(ValueError):
#             _ = repo['1,2']
#
#         # Check the doc ID must be aligned to the doc size.
#         repo = self.app.create_archived_log_repo(notification_log, doc_size)
#         with self.assertRaises(ValueError):
#             _ = repo['2,6']
#
#         # Check an archived log ID that can't be split by ',' results in a ValueError.
#         with self.assertRaises(ValueError):
#             _ = repo['invalid']
#
#     def test_archived_log_reader(self):
#         # Build a notification log (fixture).
#         notification_log = self.app.get_or_create_notification_log('log1', sequence_size=10)
#         self.append_notifications(notification_log, 13)
#
#         # Construct a feed object.
#         archived_log_repo = self.app.create_archived_log_repo(notification_log, 5)
#
#         # Use a feed reader to read the feed.
#         reader = ArchivedLogReader(archived_log_repo)
#         self.assertEqual(len(list(reader.get_items())), 13, list(reader.get_items()))
#
#         # Add some more items to the log.
#         self.append_notifications(notification_log, 13, 21)
#
#         # Use archived log reader to generate notifications from the archived log repo.
#         reader = ArchivedLogReader(archived_log_repo)
#         self.assertEqual(len(list(reader.get_items())), 21)
#
#         # Read items after last item number.
#         self.assertEqual(len(list(reader.get_items(last_item_num=1))), 20)
#         self.assertEqual(len(list(reader.get_items(last_item_num=2))), 19)
#         self.assertEqual(len(list(reader.get_items(last_item_num=3))), 18)
#         self.assertEqual(len(list(reader.get_items(last_item_num=19))), 2)
#         self.assertEqual(len(list(reader.get_items(last_item_num=20))), 1)
#         self.assertEqual(len(list(reader.get_items(last_item_num=21))), 0)
#         self.assertEqual(len(list(reader.get_items(last_item_num=22))), 0)
#         self.assertEqual(len(list(reader.get_items(last_item_num=23))), 0)
#
#         # Check last item numbers less than 1 cause a value errors.
#         with self.assertRaises(ValueError):
#             list(reader.get_items(last_item_num=-1))
#
#         with self.assertRaises(ValueError):
#             list(reader.get_items(last_item_num=0))
#
#         # Use a feed reader to read the feed with a larger doc size.
#         reader = ArchivedLogReader(archived_log_repo)
#         self.assertEqual(len(list(reader.get_items())), 21)
#
#     def test_remote_archived_log_repo(self):
#         log_name = 'log2'
#         sequence_size = 100
#         doc_size = 20
#         num_notifications = 42
#         notification_log = self.app.get_or_create_notification_log(log_name, sequence_size)
#
#         # Build a notification log (fixture).
#         self.append_notifications(notification_log, num_notifications)
#
#         # Start a simple server.
#         from wsgiref.util import setup_testing_defaults
#         from wsgiref.simple_server import make_server
#
#         port = 8000
#         base_url = 'http://127.0.0.1:{}/notifications/'.format(port)
#
#         def simple_app(environ, start_response):
#             """Simple WSGI application, for testing purposes only."""
#             setup_testing_defaults(environ)
#             status = '200 OK'
#             headers = [('Content-type', 'text/plain; charset=utf-8')]
#             start_response(status, headers)
#
#             # Extract log name and doc ID from path info.
#             path_info = environ['PATH_INFO']
#             try:
#                 notification_log_name, archived_log_id = path_info.strip('/').split('/')[-2:]
#             except ValueError as e:
#                 msg = "Couldn't extract log name and doc ID from path info {}: {}".format(path_info, e)
#                 raise ValueError(msg)
#
#             # Get serialized archived log.
#             archived_log = self.app.create_archived_log_repo(notification_log, doc_size)[archived_log_id]
#             archived_log_json = serialize_archived_log(archived_log)
#
#             # Return a list of lines.
#             return [(line + '\n').encode('utf8') for line in archived_log_json.split('\n')]
#
#         httpd = make_server('', port, simple_app)
#         print("Serving on port {}...".format(port))
#         thread = Thread(target=httpd.serve_forever)
#         thread.setDaemon(True)
#         thread.start()
#         try:
#             # Use reader with client to read all items in remote feed after item 5.
#             archived_log_repo = RemoteArchivedLogRepo(base_url, log_name)
#
#             # Get all the items.
#             archived_log_reader = ArchivedLogReader(archived_log_repo=archived_log_repo)
#             items_from_start = list(archived_log_reader.get_items())
#
#             # Check we got all the items.
#             self.assertEqual(len(items_from_start), num_notifications)
#             self.assertEqual(items_from_start[0], 'item1')
#             expected_log_count = (num_notifications // doc_size) + (1 if num_notifications % doc_size else 0)
#             self.assertEqual(archived_log_reader.archived_log_count, expected_log_count)
#
#             # Get all the items from item 5.
#             items_from_5 = list(archived_log_reader.get_items(last_item_num=doc_size))
#
#             # Check we got everything after item 5.
#             self.assertEqual(len(items_from_5), num_notifications - doc_size)
#             self.assertEqual(items_from_5[0], 'item{}'.format(doc_size + 1))
#             expected_log_count = (num_notifications // doc_size) + (1 if num_notifications % doc_size else 0) - 1
#             self.assertEqual(archived_log_reader.archived_log_count, expected_log_count)
#
#         finally:
#             httpd.shutdown()
#             thread.join()
#             httpd.server_close()
#
#     def assert_archived_log(self, repo, archived_log_id, expected_id, expected_len_items, expected_previous_id,
#                             expected_next_id):
#         archived_log = repo[archived_log_id]
#         self.assertEqual(len(archived_log.items), expected_len_items)
#         self.assertEqual(archived_log.id, expected_id)
#         self.assertEqual(archived_log.previous_id, expected_previous_id)
#         self.assertEqual(archived_log.next_id, expected_next_id)
#
#     def append_notifications(self, notification_log, *range_args):
#         for i in range(*range_args):
#             item = 'item{}'.format(i + 1)
#             self.app.append_notification(notification_log, item)
#
#
# class TestArchivedLogWithCassandra(WithCassandraActiveRecordStrategies, ArchivedLogTestCase):
#     pass
#
#
# class TestArchivedLogWithSQLAlchemy(WithSQLAlchemyActiveRecordStrategies, ArchivedLogTestCase):
#     use_named_temporary_file = True
#     # pass
#
#
# class TestErrors(TestCase):
#     def test_errors(self):
#         with self.assertRaises(ValueError):
#             deserialise_archived_log('invalid json')



import datetime
from json import JSONDecoder, JSONEncoder
from uuid import UUID

import dateutil.parser

from eventsourcing.domain.model.events import resolve_domain_topic, topic_from_domain_class


class ObjectJSONEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return {'ISO8601_datetime': obj.strftime('%Y-%m-%dT%H:%M:%S.%f%z')}
        elif isinstance(obj, datetime.date):
            return {'ISO8601_date': obj.isoformat()}
        elif isinstance(obj, UUID):
            return {'UUID': obj.hex}
        elif hasattr(obj, '__class__') and hasattr(obj, '__dict__'):
            topic = topic_from_domain_class(obj.__class__)
            state = obj.__dict__.copy()
            return {
                '__class__': {
                    'topic': topic,
                    'state': state,
                }
            }
        # Let the base class default method raise the TypeError.
        return JSONEncoder.default(self, obj)


class ObjectJSONDecoder(JSONDecoder):
    def __init__(self, object_hook=None, **kwargs):
        super(ObjectJSONDecoder, self).__init__(object_hook=object_hook or self.from_jsonable, **kwargs)

    @classmethod
    def from_jsonable(cls, d):
        if 'ISO8601_datetime' in d:
            return cls._decode_datetime(d)
        elif 'ISO8601_date' in d:
            return cls._decode_date(d)
        elif 'UUID' in d:
            return cls._decode_uuid(d)
        elif '__class__' in d:
            return cls._decode_object(d)
        return d

    @staticmethod
    def _decode_date(d):
        return datetime.datetime.strptime(d['ISO8601_date'], '%Y-%m-%d').date()

    @staticmethod
    def _decode_datetime(d):
        return dateutil.parser.parse(d['ISO8601_datetime'])

    @staticmethod
    def _decode_uuid(d):
        return UUID(d['UUID'])

    @staticmethod
    def _decode_object(d):
        topic = d['__class__']['topic']
        state = d['__class__']['state']
        obj_class = resolve_domain_topic(topic)
        obj = object.__new__(obj_class)
        obj.__dict__.update(state)
        return obj



from eventsourcing.application.base import EventSourcedApplication
from eventsourcing.example.domainmodel import register_new_example
from eventsourcing.example.infrastructure import ExampleRepository
from eventsourcing.infrastructure.snapshotting import EventSourcedSnapshotStrategy


class ExampleApplication(EventSourcedApplication):
    """
    Abstract example event sourced application.

    This application has an Example repository, and a factory method to construct new Example entities.

    It doesn't have a stored event repository.
    """

    def __init__(self, **kwargs):
        super(ExampleApplication, self).__init__(**kwargs)
        self.snapshot_strategy = EventSourcedSnapshotStrategy(
            event_store=self.timestamp_entity_event_store,
        )
        self.example_repo = ExampleRepository(
            event_store=self.version_entity_event_store,
            snapshot_strategy=self.snapshot_strategy,
        )

    def register_new_example(self, a, b):
        return register_new_example(a=a, b=b)



from setuptools import setup, find_packages

try:
    from functools import singledispatch
    install_requires_singledispatch = []
except ImportError:
    install_requires_singledispatch = ['singledispatch']


setup(
    name='eventsourcing',
    version='2.1.1',
    description='Event sourcing in Python',
    author='John Bywater',
    author_email='john.bywater@appropriatesoftware.net',
    url='https://github.com/johnbywater/eventsourcing',
    packages=find_packages(),
    install_requires=[
        'python-dateutil',
        'singledispatch',
        'six',
    ] + install_requires_singledispatch,
    extras_require={
        'cassandra': [
            'cassandra-driver',
        ],
        'crypto': [
            'PyCrypto',
        ],
        'test': [
            'cassandra-driver',
            'mock',
            'PyCrypto',
            'sqlalchemy',
            'sqlalchemy-utils',
            'requests',
        ],
        'sqlalchemy': [
            'sqlalchemy',
            'sqlalchemy-utils',
        ],
    },
    zip_safe=False,
    long_description="""
This package provides generic support for event sourcing in Python.

An extensive
`README file is available on GitHub <https://github.com/johnbywater/eventsourcing/blob/master/README.md>`_.
""",
    keywords=['event sourcing', 'event store', 'domain driven design', 'ddd', 'cqrs', 'cqs'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



from unittest.case import TestCase
from uuid import uuid4

from cassandra import InvalidRequest
from cassandra.cluster import NoHostAvailable
from cassandra.cqlengine import CQLEngineException

from eventsourcing.exceptions import DatasourceSettingsError
from eventsourcing.infrastructure.cassandra.activerecords import CqlIntegerSequencedItem, CqlTimestampSequencedItem
from eventsourcing.infrastructure.cassandra.datastore import CassandraDatastore, CassandraSettings
from eventsourcing.infrastructure.datastore import DatastoreConnectionError, DatastoreTableError
from eventsourcing.tests.datastore_tests.base import AbstractDatastoreTestCase, DatastoreTestCase

DEFAULT_KEYSPACE_FOR_TESTING = 'eventsourcing_tests'


class CassandraDatastoreTestCase(AbstractDatastoreTestCase):
    """
    Uses the datastore object to set up connection to and tables in Cassandra.
    """

    def construct_datastore(self):
        return CassandraDatastore(
            settings=CassandraSettings(default_keyspace=DEFAULT_KEYSPACE_FOR_TESTING),
            tables=(CqlIntegerSequencedItem, CqlTimestampSequencedItem),
        )


class TestCassandraDatastore(CassandraDatastoreTestCase, DatastoreTestCase):
    def list_records(self):
        try:
            return list(CqlIntegerSequencedItem.objects.all())
        except (CQLEngineException, NoHostAvailable) as e:
            raise DatastoreConnectionError(e)
        except InvalidRequest as e:
            raise DatastoreTableError(e)

    def create_record(self):
        record = CqlIntegerSequencedItem(s=uuid4(), p=0, t='topic', d='{}')
        try:
            record.save()
        except (CQLEngineException, NoHostAvailable) as e:
            raise DatastoreConnectionError(e)
        except InvalidRequest as e:
            raise DatastoreTableError(e)
        return record


class TestPlainTextAuthProvider(TestCase):
    def test_plain_text_auth_provider(self):
        datastore = CassandraDatastore(
            settings=CassandraSettings(
                default_keyspace=DEFAULT_KEYSPACE_FOR_TESTING,
                username='username',
                password='password',
            ),
            tables=(CqlIntegerSequencedItem,),
        )
        datastore.setup_connection()


class TestDatabaseSettingsError(TestCase):
    def test_consistency_level_error(self):
        datastore = CassandraDatastore(
            settings=CassandraSettings(
                default_keyspace=DEFAULT_KEYSPACE_FOR_TESTING,
                consistency='invalid',
            ),
            tables=(CqlIntegerSequencedItem,),
        )

        with self.assertRaises(DatasourceSettingsError):
            datastore.setup_connection()



from uuid import uuid4

import mock

from eventsourcing.domain.model.entity import AttributeChanged, Created, CreatedMutatorRequiresTypeNotInstance, \
    EntityIDConsistencyError, EntityVersionConsistencyError, TimestampedVersionedEntity, attribute, created_mutator
from eventsourcing.domain.model.events import VersionedEntityEvent, publish, subscribe, unsubscribe, DomainEvent
from eventsourcing.example.infrastructure import ExampleRepository
from eventsourcing.example.domainmodel import Example, register_new_example
from eventsourcing.exceptions import ProgrammingError, RepositoryKeyError, ConcurrencyError
from eventsourcing.tests.sequenced_item_tests.base import WithPersistencePolicy
from eventsourcing.tests.sequenced_item_tests.test_cassandra_active_record_strategy import \
    WithCassandraActiveRecordStrategies
from eventsourcing.tests.sequenced_item_tests.test_sqlalchemy_active_record_strategy import \
    WithSQLAlchemyActiveRecordStrategies


class TestExampleEntity(WithSQLAlchemyActiveRecordStrategies, WithPersistencePolicy):
    def test_entity_lifecycle(self):
        # Check the factory creates an instance.
        example1 = register_new_example(a=1, b=2)
        self.assertIsInstance(example1, Example)

        # Check the properties of the Example class.
        self.assertEqual(1, example1.a)
        self.assertEqual(2, example1.b)

        # Check the properties of the TimestampedVersionedEntity class.
        self.assertTrue(example1.id)
        self.assertEqual(1, example1.version)
        self.assertTrue(example1.created_on)

        # Check a second instance with the same values is not "equal" to the first.
        example2 = register_new_example(a=1, b=2)
        self.assertNotEqual(example1, example2)

        # Setup the repo.
        repo = ExampleRepository(self.versioned_entity_event_store)

        # Check the example entities can be retrieved from the example repository.
        entity1 = repo[example1.id]
        self.assertIsInstance(entity1, Example)
        self.assertEqual(1, entity1.a)
        self.assertEqual(2, entity1.b)

        entity2 = repo[example2.id]
        self.assertIsInstance(entity2, Example)
        self.assertEqual(1, entity2.a)
        self.assertEqual(2, entity2.b)

        # Check the entity can be updated.
        entity1.a = 100
        self.assertEqual(100, repo[entity1.id].a)
        entity1.b = -200
        self.assertEqual(-200, repo[entity1.id].b)

        self.assertEqual(0, entity1.count_heartbeats())
        entity1.beat_heart()
        entity1.beat_heart()
        entity1.beat_heart()
        self.assertEqual(3, entity1.count_heartbeats())
        self.assertEqual(3, repo[entity1.id].count_heartbeats())

        # Check the entity can publish multiple events simultaneously.
        entity1.beat_heart(number_of_beats=3)
        self.assertEqual(6, repo[entity1.id].count_heartbeats())

        # Check the entity can be discarded.
        entity1.discard()

        # Check the repo now raises a KeyError.
        self.assertRaises(RepositoryKeyError, repo.__getitem__, entity1.id)

        # Check the entity can't be discarded twice.
        self.assertRaises(AssertionError, entity1.discard)

        # Should fail to validate event with wrong entity ID.
        with self.assertRaises(EntityIDConsistencyError):
            entity2._validate_originator(
                VersionedEntityEvent(
                    entity_id=uuid4(),
                    entity_version=0
                )
            )
        # Should fail to validate event with wrong entity version.
        with self.assertRaises(EntityVersionConsistencyError):
            entity2._validate_originator(
                VersionedEntityEvent(
                    entity_id=entity2.id,
                    entity_version=0,
                )
        )

        # Should validate event with correct entity ID and version.
        entity2._validate_originator(
            VersionedEntityEvent(
                entity_id=entity2.id,
                entity_version=entity2.version,
            )
        )

        # Check an entity cannot be reregistered with the ID of a discarded entity.
        replacement_event = Example.Created(entity_id=entity1.id, a=11, b=12)
        with self.assertRaises(ConcurrencyError):
            publish(event=replacement_event)

    def test_not_implemented_error(self):
        # Define an event class.
        class UnsupportedEvent(DomainEvent): pass

        # Check we get an error when attempting to mutate on the event.
        self.assertRaises(NotImplementedError, Example.mutate, Example, UnsupportedEvent())

    def test_attribute(self):
        # Check we get an error when called with something other than a function.
        self.assertRaises(ProgrammingError, attribute, 'not a getter')
        self.assertRaises(ProgrammingError, attribute, 123)
        self.assertRaises(ProgrammingError, attribute, None)

        # Call the decorator with a function.
        getter = lambda: None
        p = attribute(getter)

        # Check we got a property object.
        self.assertIsInstance(p, property)

        # Check the property object has both setter and getter functions.
        self.assertTrue(p.fset)
        self.assertTrue(p.fget)

        # Pretend we decorated an object.
        entity_id = uuid4()
        o = TimestampedVersionedEntity(entity_id=entity_id, entity_version=0)
        o.__dict__['_<lambda>'] = 'value1'

        # Call the property's getter function.
        value = p.fget(o)
        self.assertEqual(value, 'value1')

        # Call the property's setter function.
        p.fset(o, 'value2')

        # Check the attribute has changed.
        value = p.fget(o)
        self.assertEqual(value, 'value2')

        # Check the property's getter function isn't the getter function we passed in.
        self.assertNotEqual(p.fget, getter)

        # Define a class that uses the decorator.
        class Aaa(TimestampedVersionedEntity):
            "An event sourced entity."

            def __init__(self, a, *args, **kwargs):
                super(Aaa, self).__init__(*args, **kwargs)
                self._a = a

            @attribute
            def a(self):
                "A mutable event sourced property."

        # Instantiate the class and check assigning to the property publishes an event and updates the object state.
        published_events = []
        subscription = (lambda x: True, lambda x: published_events.append(x))
        subscribe(*subscription)
        entity_id = uuid4()
        try:
            aaa = Aaa(entity_id=entity_id, entity_version=1, a=1)
            self.assertEqual(aaa.a, 1)
            aaa.a = 'value1'
            self.assertEqual(aaa.a, 'value1')
        finally:
            unsubscribe(*subscription)

        # Check an event was published.
        self.assertEqual(len(published_events), 1)

        # Check the published event was an AttributeChanged event, with the expected attribute values.
        published_event = published_events[0]
        self.assertIsInstance(published_event, AttributeChanged)
        self.assertEqual(published_event.name, '_a')
        self.assertEqual(published_event.value, 'value1')
        self.assertTrue(published_event.entity_version, 1)
        self.assertEqual(published_event.entity_id, entity_id)

    def test_static_mutator_method(self):
        self.assertRaises(NotImplementedError, TimestampedVersionedEntity._mutator, 1, 2)

    def test_created_mutator_errors(self):
        # Check the guard condition raises exception.
        with self.assertRaises(CreatedMutatorRequiresTypeNotInstance):
            created_mutator(mock.Mock(spec=Created), 'not a class')

        # Check the instantiation type error.
        with self.assertRaises(TypeError):
            created_mutator(mock.Mock(spec=Created), TimestampedVersionedEntity)


class CustomValueObject(object):
    def __init__(self, value):
        self.value = value


class TestExampleEntityWithCassandra(WithCassandraActiveRecordStrategies, TestExampleEntity):
    pass



# from eventsourcing.domain.model.notificationlog import NotificationLog, NotificationLogRepository
# from eventsourcing.infrastructure.eventsourcedrepository import EventSourcedRepository
#
#
# class NotificationLogRepo(EventSourcedRepository, NotificationLogRepository):
#     """
#     Event sourced repository for the Example domain model entity.
#     """
#     domain_class = NotificationLog



# from eventsourcing.example.application import ExampleApplication
# from eventsourcing.example.domainmodel import Example
# from eventsourcing.exceptions import ConcurrencyError, SequencedItemError
# from eventsourcing.tests.example_application_tests.base import ExampleApplicationTestCase
# from eventsourcing.tests.sequenced_item_tests.test_sqlalchemy_active_record_strategy import \
#     WithSQLAlchemyActiveRecordStrategies
#
#
# class TestFastForward(WithSQLAlchemyActiveRecordStrategies, ExampleApplicationTestCase):
#
#     def test(self):
#         with self.construct_application() as app:
#             assert isinstance(app, ExampleApplication)
#             example = app.register_new_example(1, 2)
#             instance1 = app.example_repo[example.id]
#             instance2 = app.example_repo[example.id]
#
#             assert isinstance(instance1, Example)
#             assert isinstance(instance2, Example)
#
#             self.assertEqual(instance1.version, 1)
#             self.assertEqual(instance2.version, 1)
#
#             # Evolve instance1 by a version.
#             instance1.beat_heart()
#             self.assertEqual(instance1.version, 2)
#
#             # Fail to evolve instance2 in the same way.
#             # Todo: This needs to be a deepcopy.
#             preop_state = instance2.__dict__.copy()
#             with self.assertRaises(SequencedItemError):
#                 instance2.beat_heart()
#
#             # Reset instance2 to its pre-op state.
#             instance2.__dict__.update(preop_state)
#             self.assertEqual(instance2.version, 1)
#
#             # Fast forward instance2 from pre-op state.
#             instance3 = app.example_repo.fastforward(instance2)
#             self.assertEqual(instance2.version, 1)
#             self.assertEqual(instance3.version, 2)
#
#             # Try again to beat heart.
#             instance3.beat_heart()
#             self.assertEqual(instance3.version, 3)
#
#             # Try to evolve instance1 from its stale version.
#             preop_state = instance1.__dict__.copy()
#             with self.assertRaises(ConcurrencyError):
#                 instance1.beat_heart()
#
#             # Reset instance1 to pre-op state.
#             instance1.__dict__.update(preop_state)
#
#             # Fast forward instance1 from pre-op state.
#             instance4 = app.example_repo.fastforward(instance1)
#             self.assertEqual(instance1.version, 2)
#             self.assertEqual(instance4.version, 3)
#
#             # Try again to beat heart.
#             instance4.beat_heart()
#             self.assertEqual(instance4.version, 4)



from uuid import uuid4

from eventsourcing.application.policies import CombinedPersistencePolicy
from eventsourcing.domain.model.events import assert_event_handlers_empty
from eventsourcing.example.domainmodel import Example, register_new_example
from eventsourcing.infrastructure.eventplayer import EventPlayer
from eventsourcing.infrastructure.eventstore import EventStore
from eventsourcing.infrastructure.sequenceditem import SequencedItem
from eventsourcing.infrastructure.snapshotting import entity_from_snapshot, EventSourcedSnapshotStrategy
from eventsourcing.infrastructure.sqlalchemy.activerecords import SQLAlchemyActiveRecordStrategy, \
    SqlIntegerSequencedItem, SqlTimestampSequencedItem
from eventsourcing.infrastructure.sequenceditemmapper import SequencedItemMapper
from eventsourcing.tests.datastore_tests.test_sqlalchemy import SQLAlchemyDatastoreTestCase


class TestEventPlayer(SQLAlchemyDatastoreTestCase):
    def setUp(self):
        assert_event_handlers_empty()
        super(TestEventPlayer, self).setUp()

        self.datastore.setup_connection()
        self.datastore.setup_tables()

        # Setup an event store for version entity events.
        self.version_entity_event_store = EventStore(
            active_record_strategy=SQLAlchemyActiveRecordStrategy(
                datastore=self.datastore,
                active_record_class=SqlIntegerSequencedItem,
                sequenced_item_class=SequencedItem,
            ),
            sequenced_item_mapper=SequencedItemMapper(
                sequenced_item_class=SequencedItem,
                event_sequence_id_attr='entity_id',
                event_position_attr='entity_version',
            ),
        )

        # Setup an event store for timestamp entity events.
        self.timestamp_entity_event_store = EventStore(
            active_record_strategy=SQLAlchemyActiveRecordStrategy(
                datastore=self.datastore,
                active_record_class=SqlTimestampSequencedItem,
                sequenced_item_class=SequencedItem,
            ),
            sequenced_item_mapper=SequencedItemMapper(
                sequenced_item_class=SequencedItem,
                event_sequence_id_attr='entity_id',
                event_position_attr='timestamp',
            ),
        )

        self.policy = None

    def tearDown(self):
        self.datastore.drop_tables()
        self.datastore.drop_connection()
        if self.policy is not None:
            self.policy.close()
        super(TestEventPlayer, self).tearDown()
        assert_event_handlers_empty()

    def test_get_entity(self):
        # Store example events.
        entity_id1 = uuid4()
        event1 = Example.Created(entity_id=entity_id1, a=1, b=2)
        self.version_entity_event_store.append(event1)
        entity_id2 = uuid4()
        event2 = Example.Created(entity_id=entity_id2, a=2, b=4)
        self.version_entity_event_store.append(event2)
        entity_id3 = uuid4()
        event3 = Example.Created(entity_id=entity_id3, a=3, b=6)
        self.version_entity_event_store.append(event3)
        event4 = Example.Discarded(entity_id=entity_id3, entity_version=1)
        self.version_entity_event_store.append(event4)

        # Check the event sourced entities are correct.
        # - just use a trivial mutate that always instantiates the 'Example'.
        event_player = EventPlayer(event_store=self.version_entity_event_store, mutator=Example.mutate)

        # The the reconstituted entity has correct attribute values.
        self.assertEqual(entity_id1, event_player.replay_entity(entity_id1).id)
        self.assertEqual(1, event_player.replay_entity(entity_id1).a)
        self.assertEqual(2, event_player.replay_entity(entity_id2).a)
        self.assertEqual(None, event_player.replay_entity(entity_id3))

        # Check entity3 raises KeyError.
        self.assertEqual(event_player.replay_entity(entity_id3), None)

        # Check it works for "short" entities (should be faster, but the main thing is that it still works).
        # - just use a trivial mutate that always instantiates the 'Example'.
        event5 = Example.AttributeChanged(entity_id=entity_id1, entity_version=1, name='a', value=10)
        self.version_entity_event_store.append(event5)

        event_player = EventPlayer(event_store=self.version_entity_event_store, mutator=Example.mutate)
        self.assertEqual(10, event_player.replay_entity(entity_id1).a)

        event_player = EventPlayer(
            event_store=self.version_entity_event_store,
            mutator=Example.mutate,
            is_short=True,
        )
        self.assertEqual(10, event_player.replay_entity(entity_id1).a)

    # Todo: Maybe this is an application-level test? If not, test event player capabilities here only.
    def test_snapshots(self):
        self.policy = CombinedPersistencePolicy(
            versioned_entity_event_store=self.version_entity_event_store,
            timestamped_entity_event_store=self.timestamp_entity_event_store,

        )
        event_player = EventPlayer(
            event_store=self.version_entity_event_store,
            mutator=Example.mutate,
            snapshot_strategy=EventSourcedSnapshotStrategy(
                event_store=self.timestamp_entity_event_store
            )
        )

        # Take a snapshot with a non-existent ID.
        self.assertIsNone(event_player.take_snapshot(uuid4()))

        # Create a new entity.
        registered_example = register_new_example(a=123, b=234)

        # Take a snapshot.
        snapshot1 = event_player.take_snapshot(registered_example.id)

        # Replay from this snapshot.
        initial_state = entity_from_snapshot(snapshot1)
        retrieved_example = event_player.replay_entity(registered_example.id,
                                                       initial_state=initial_state,
                                                       gte=initial_state._version)

        # Check the attributes are correct.
        self.assertEqual(retrieved_example.a, 123)

        # Remember the version now.
        version1 = retrieved_example._version

        # Change attribute value.
        retrieved_example.a = 999

        # Remember the version now.
        version2 = retrieved_example._version

        # Change attribute value.
        retrieved_example.a = 9999

        # Remember the version now.
        version3 = retrieved_example._version

        # Check the event sourced entities are correct.
        retrieved_example = event_player.replay_entity(registered_example.id)
        self.assertEqual(retrieved_example.a, 9999)

        # Take another snapshot.
        snapshot2 = event_player.take_snapshot(retrieved_example.id)

        # Check we can replay from this snapshot.
        initial_state = entity_from_snapshot(snapshot2)
        retrieved_example = event_player.replay_entity(
            registered_example.id,
            initial_state=initial_state,
            gte=initial_state._version,
        )
        # Check the attributes are correct.
        self.assertEqual(retrieved_example.a, 9999)

        # Check we can get historical state at version1.
        retrieved_example = event_player.replay_entity(registered_example.id, lt=version1)
        self.assertEqual(retrieved_example.a, 123)

        # Check we can get historical state at version2.
        retrieved_example = event_player.replay_entity(registered_example.id, lt=version2)
        self.assertEqual(retrieved_example.a, 999)

        # Check we can get historical state at version3.
        retrieved_example = event_player.replay_entity(registered_example.id, lt=version3)
        self.assertEqual(retrieved_example.a, 9999)

        # Similarly, check we can get historical state using a snapshot
        initial_state = entity_from_snapshot(snapshot1)
        retrieved_example = event_player.replay_entity(
            registered_example.id,
            initial_state=initial_state,
            gte=initial_state._version,
            lt=version2,
        )
        self.assertEqual(retrieved_example.a, 999)
